import pytest
from datetime import datetime, timedelta

from .timewindow import RelAbsTime, TimeWindow
from freezegun import freeze_time


class TestRelAbs:
    def test_defaults(self):
        rel_time = RelAbsTime(None)
        assert rel_time.is_after()
        assert rel_time.is_before()

        cur_date = datetime.now()

        rel_time = RelAbsTime(cur_date.strftime("%H:%M"))
        assert rel_time.is_after()
        assert not rel_time.is_before()

        rel_time = RelAbsTime(None, default_time=cur_date)
        assert rel_time.is_after()
        assert not rel_time.is_before()
        assert rel_time.offset_s is None

    def test_offset(self):
        cur_date = datetime.now()
        rel_time = RelAbsTime("30m", default_time=cur_date)
        assert rel_time.offset_s == 30 * 60
        rel_time = RelAbsTime("30s", default_offset=200)
        assert rel_time.offset_s == 30
        rel_time = RelAbsTime("1.5h")
        assert rel_time.offset_s == 1.5 * 60 * 60
        rel_time = RelAbsTime("1a.d5h", default_time=cur_date)
        assert rel_time.offset_s is None
        assert rel_time.time == cur_date.time()

        rel_time = RelAbsTime("3.5")
        assert rel_time.offset_s == 3.5 * 60
        rel_time = RelAbsTime("3.5z")
        assert rel_time.offset_s == 3.5


class TestWindow:
    # christchurch
    DEFAULT_LAT = -43.5321
    DEFAULT_LONG = 172.6362

    @freeze_time(lambda: datetime.now().replace(hour=1, minute=59))
    def test_before_sunrise(self):
        start = RelAbsTime("-30m")
        end = RelAbsTime("30m")
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        assert time_window.inside_window()

    @freeze_time(lambda: datetime.now().replace(hour=23, minute=59))
    def test_after_sunset(self):
        start = RelAbsTime("-30m")
        end = RelAbsTime("30m")
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        assert time_window.inside_window()

    @freeze_time(
        lambda: (datetime.now() - timedelta(days=1)).replace(hour=23, minute=59)
    )
    def test_after_midnight(self):
        start = RelAbsTime("-30m")
        end = RelAbsTime("30m")
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        assert time_window.inside_window()
        with freeze_time(
            (datetime.now() + timedelta(days=1)).replace(hour=1, minute=59)
        ):
            assert time_window.inside_window()

    @freeze_time(lambda: datetime.now().replace(hour=1, minute=59))
    def test_dt_update(self):
        start = RelAbsTime("-30m")
        end = RelAbsTime("30m")
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        assert time_window.inside_window()
        end_date = time_window.end.dt
        start_date = time_window.start.dt

        # check that if we move out of window, the end time and start time gets updated
        with freeze_time(datetime.now().replace(hour=12, minute=59)):
            time_window.inside_window()

            assert end_date != time_window.end.dt
            assert start_date != time_window.start.dt

    @freeze_time(lambda: datetime.now().replace(hour=12, minute=1))
    def test_after_sunrise(self):
        start = RelAbsTime("-30m")
        end = RelAbsTime("30m")
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        assert not time_window.inside_window()

    def test_absolute_times(self):
        cur_date = datetime.now()
        start = RelAbsTime(cur_date.strftime("%H:%M"))
        end = RelAbsTime(cur_date.strftime("%H:%M"))
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        assert time_window.inside_window()

        new_end = cur_date + timedelta(minutes=1)
        if new_end.day < cur_date.day:
            new_end = cur_date
        start = RelAbsTime(cur_date.strftime("%H:%M"))
        end = RelAbsTime(new_end.strftime("%H:%M"))
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )

        assert time_window.inside_window()

        new_end = cur_date + timedelta(minutes=-2)
        if new_end.day < cur_date.day:
            new_end = cur_date
        start = RelAbsTime(cur_date.strftime("%H:%M"))
        end = RelAbsTime(new_end.strftime("%H:%M"))

        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )

        assert time_window.inside_window()

    @freeze_time(lambda: datetime.now().replace(hour=12, minute=30))
    def test_sunrise_times(self):
        cur_date = datetime.now()
        start = RelAbsTime(cur_date.strftime("%H:%M"))
        end = RelAbsTime("0s")
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        time_window.inside_window()
        assert time_window.last_sunrise_check is not None

        time_window.end.dt = cur_date + timedelta(hours=11)
        assert time_window.inside_window()

        start = RelAbsTime(cur_date.strftime("0s"))
        end = RelAbsTime("0s")
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        assert time_window.start.dt.date() == datetime.now().date()
        assert time_window.end.dt > time_window.start.dt
        assert time_window.end.dt.date() == datetime.now().date() + timedelta(days=1)
        prev_s = time_window.start.dt.date()
        time_window.update_sun_times()
        assert time_window.start.dt.date() == prev_s

        time_window.update_sun_times(True)
        assert time_window.start.dt.date() == prev_s + timedelta(days=1)

        time_window.inside_window()
        assert time_window.start.dt.date() == prev_s + timedelta(days=1)

        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        prev_s = time_window.start.dt.date()
        time_window.inside_window()

        time_window.start.dt = time_window.start.dt - timedelta(days=1)
        time_window.end.dt = time_window.end.dt - timedelta(days=1)
        prev_s = time_window.start.dt.date()

        # current time is after start and end so sould pick tomorrows sunset window
        in_w = time_window.inside_window()
        assert prev_s != time_window.start.dt.date()

    @freeze_time(lambda: datetime.now().replace(hour=17, minute=2))
    def test_absolute_times_days(self):
        start = RelAbsTime("5:00")
        end = RelAbsTime("08:00")
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        assert not time_window.inside_window()
        assert start.dt > datetime.now()
        assert end.dt > datetime.now()

        start = RelAbsTime("17:00")
        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        assert time_window.inside_window()
        assert start.dt < datetime.now()
        assert end.dt.date() == (datetime.now() + timedelta(days=1)).date()

    @freeze_time(lambda: datetime.now().replace(hour=17, minute=2))
    def test_absolute_and_relative(self):
        start = RelAbsTime("17:00")
        end = RelAbsTime("+30m")

        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        time_window.update_sun_times()

        assert time_window.inside_window()
        assert start.dt < end.dt

    @freeze_time(lambda: datetime.now().replace(hour=10, minute=18))
    def test_constant(self):
        start = RelAbsTime("12:00")
        end = RelAbsTime("12:00")

        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
        )
        assert time_window.inside_window()
        assert start.dt <= end.dt
        with freeze_time((datetime.now() + timedelta(hours=10))):
            time_window = TimeWindow(
                start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
            )
            assert time_window.inside_window()

    @freeze_time(lambda: datetime.now().replace(hour=1, minute=2))
    def test_in_window(self):
        start = RelAbsTime("17:00")
        end = RelAbsTime("8:00")

        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, None
        )
        assert time_window.inside_window()
        assert start.dt < end.dt

        start = RelAbsTime("17:00")
        end = RelAbsTime("+30m")

        time_window = TimeWindow(
            start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, None
        )
        assert time_window.inside_window()
        assert start.dt < end.dt

        with freeze_time(datetime.now().replace(hour=11, minute=2)):
            time_window = TimeWindow(
                start, end, TestWindow.DEFAULT_LAT, TestWindow.DEFAULT_LONG, 0
            )

            assert not time_window.inside_window()



from abc import ABC, abstractmethod
import time
import json
import logging
import numpy as np
from ml_tools.hyperparams import HyperParams
from pathlib import Path
from classify.trackprediction import TrackPrediction


class Interpreter(ABC):
    def __init__(self, model_file):
        self.load_json(model_file)
        self.id = None

    def load_json(self, filename):
        """Loads model and parameters from file."""
        filename = Path(filename)
        filename = filename.with_suffix(".json")
        logging.info("Loading metadata from %s", filename)
        metadata = json.load(open(filename, "r"))
        self.version = metadata.get("version", None)
        self.labels = metadata["labels"]
        self.params = HyperParams()
        self.params.update(metadata.get("hyperparams", {}))
        self.data_type = metadata.get("type", "thermal")

        self.mapped_labels = metadata.get("mapped_labels")
        self.label_probabilities = metadata.get("label_probabilities")
        self.preprocess_fn = self.get_preprocess_fn()

    @abstractmethod
    def shape(self):
        """Num Inputs, Prediction shape"""
        ...

    @abstractmethod
    def predict(self, frames):
        """predict"""
        ...

    def get_preprocess_fn(self):
        model_name = self.params.model_name
        if model_name == "inceptionv3":
            # no need to use tf module, if train other model types may have to add
            #  preprocess definitions
            return inc3_preprocess
        elif model_name == "wr-resnet":
            return None
        else:
            import tensorflow as tf

            if model_name == "resnet":
                return tf.keras.applications.resnet.preprocess_input
            elif model_name == "nasnet":
                return tf.keras.applications.nasnet.preprocess_input
            elif model_name == "resnetv2":
                return tf.keras.applications.resnet_v2.preprocess_input

            elif model_name == "resnet152":
                return tf.keras.applications.resnet.preprocess_input

            elif model_name == "vgg16":
                return tf.keras.applications.vgg16.preprocess_input

            elif model_name == "vgg19":
                return tf.keras.applications.vgg19.preprocess_input

            elif model_name == "mobilenet":
                return tf.keras.applications.mobilenet_v2.preprocess_input

            elif model_name == "densenet121":
                return tf.keras.applications.densenet.preprocess_input

            elif model_name == "inceptionresnetv2":
                return tf.keras.applications.inception_resnet_v2.preprocess_input
        logging.warn("pretrained model %s has no preprocessing function", model_name)
        return None

    def preprocess(self, clip, track, **args):
        scale = args.get("scale", None)
        num_predictions = args.get("num_predictions", None)
        predict_from_last = args.get("predict_from_last", None)
        segment_frames = args.get("segment_frames")
        frames_per_classify = args.get("frames_per_classify", 25)
        available_frames = (
            min(len(track.bounds_history), clip.frames_kept())
            if clip.frames_kept() is not None
            else len(track.bounds_history)
        )
        if predict_from_last is not None:
            predict_from_last = min(predict_from_last, available_frames)
        # this might be a little slower as it checks some massess etc
        # but keeps it the same for all ways of classifying
        if frames_per_classify > 1:
            if predict_from_last is not None and segment_frames is None:
                logging.debug(
                    "Prediction from last available frames %s track is of length %s",
                    available_frames,
                    len(track.bounds_history),
                )
                regions = track.bounds_history[-available_frames:]
                valid_regions = 0
                if available_frames > predict_from_last:
                    # want to get rid of any blank frames
                    predict_from_last = 0
                    for i, r in enumerate(
                        reversed(track.bounds_history[-available_frames:])
                    ):
                        if r.blank:
                            continue
                        valid_regions += 1
                        predict_from_last = i + 1
                        if valid_regions >= predict_from_last:
                            break
                logging.debug(
                    "After checking blanks have predict from last %s from last available frames %s track is of length %s",
                    predict_from_last,
                    available_frames,
                    len(track.bounds_history),
                )

            frames, preprocessed, masses = self.preprocess_segments(
                clip,
                track,
                num_predictions,
                predict_from_last,
                segment_frames=segment_frames,
                dont_filter=args.get("dont_filter", False),
                min_segments=args.get("min_segments"),
            )
        else:
            frames, preprocessed, masses = self.preprocess_frames(
                clip, track, num_predictions, segment_frames=segment_frames
            )
        return frames, preprocessed, masses

    def classify_track(self, clip, track, segment_frames=None, min_segments=None):
        start = time.time()
        prediction_frames, output, masses = self.predict_track(
            clip,
            track,
            segment_frames=segment_frames,
            frames_per_classify=self.params.square_width**2,
            min_segments=min_segments,
        )
        if output is None:
            logging.info("Skipping track %s", track.get_id())
            return None
        track_prediction = TrackPrediction(track.get_id(), self.labels)
        # self.model.predict(preprocessed)
        top_score = None
        smoothed_predictions = None

        if self.params.smooth_predictions:
            masses = np.array(masses)
            top_score = np.sum(masses)
            masses = masses[:, None]
            smoothed_predictions = output * masses
        track_prediction.classified_clip(
            output,
            smoothed_predictions,
            prediction_frames,
            masses,
            top_score=top_score,
        )
        if (
            len(prediction_frames) == 1
            and len(set(prediction_frames[0])) < self.params.square_width**2 / 4
        ):
            # if we don't have many frames to get a good prediction, lets assume only false-positive is a good prediction and filter the rest to a maximum of 0.5
            if track_prediction.predicted_tag() != "false-positive":
                track_prediction.cap_confidences(0.5)
        track_prediction.classify_time = time.time() - start
        return track_prediction

    def predict_track(self, clip, track, **args):
        frames, preprocessed, masses = self.preprocess(clip, track, **args)
        if preprocessed is None or len(preprocessed) == 0:
            return None, None, None
        pred = self.predict(preprocessed)
        return frames, pred, masses

    def preprocess_frames(
        self,
        clip,
        track,
        max_frames=None,
        segment_frames=None,
    ):
        from ml_tools.preprocess import preprocess_single_frame, preprocess_frame

        data = []
        frames_used = []
        filtered_norm_limits = None
        thermal_norm_limits = None
        if self.params.diff_norm or self.params.thermal_diff_norm:
            min_diff = None
            max_diff = 0
            thermal_max_diff = None
            thermal_min_diff = None
            for i, region in enumerate(reversed(track.bounds_history)):
                if region.blank:
                    continue
                if region.width == 0 or region.height == 0:
                    logging.warn(
                        "No width or height for frame %s regoin %s",
                        region.frame_number,
                        region,
                    )
                    continue
                f = clip.get_frame(region.frame_number)
                if region.blank or region.width <= 0 or region.height <= 0:
                    continue

                f.float_arrays()

                if self.params.thermal_diff_norm:
                    diff_frame = f.thermal - np.median(f.thermal)
                    new_max = np.amax(diff_frame)
                    new_min = np.amin(diff_frame)
                    if thermal_min_diff is None or new_min < thermal_min_diff:
                        thermal_min_diff = new_min
                    if thermal_max_diff is None or new_max > thermal_max_diff:
                        thermal_max_diff = new_max
                if self.params.diff_norm:
                    diff_frame = region.subimage(f.filtered)
                    # region.subimage(f.thermal) - region.subimage(
                    # clip.background
                    # )

                    new_max = np.amax(diff_frame)
                    new_min = np.amin(diff_frame)
                    if min_diff is None or new_min < min_diff:
                        min_diff = new_min
                    if new_max > max_diff:
                        max_diff = new_max
            if self.params.thermal_diff_norm:
                thermal_norm_limits = (thermal_min_diff, thermal_max_diff)

            if self.params.diff_norm:
                filtered_norm_limits = (min_diff, max_diff)

        for i, region in enumerate(reversed(track.bounds_history)):
            if region.blank:
                continue
            if region.width == 0 or region.height == 0:
                logging.warn(
                    "No width or height for frame %s regoin %s",
                    region.frame_number,
                    region,
                )
                continue
            frame = clip.get_frame(region.frame_number)
            if frame is None:
                logging.error(
                    "Clasifying clip %s track %s can't get frame %s",
                    clip.get_id(),
                    track.get_id(),
                    region.frame_number,
                )
                raise Exception(
                    "Clasifying clip {} track {} can't get frame {}".format(
                        clip.get_id(), track.get_id(), region.frame_number
                    )
                )
            logging.debug(
                "classifying single frame with preprocess %s size %s crop? %s f shape %s region %s",
                "None" if self.preprocess_fn is None else self.preprocess_fn.__module__,
                self.params.frame_size,
                True,
                frame.thermal.shape,
                region,
            )
            cropped_frame = preprocess_frame(
                frame,
                (self.params.frame_size, self.params.frame_size),
                region,
                clip.background,
                clip.crop_rectangle,
                calculate_filtered=False,
                filtered_norm_limits=filtered_norm_limits,
                thermal_norm_limits=thermal_norm_limits,
            )
            preprocessed = preprocess_single_frame(
                cropped_frame,
                self.params.channels,
                self.preprocess_fn,
                save_info=f"{region.frame_number} - {region}",
            )

            frames_used.append(region.frame_number)
            data.append(preprocessed)
            if max_frames is not None and len(data) >= max_frames:
                break
        return frames_used, np.array(data), [region.mass]

    def preprocess_segments(
        self,
        clip,
        track,
        max_segments=None,
        predict_from_last=None,
        segment_frames=None,
        dont_filter=False,
        min_segments=None,
    ):
        from ml_tools.preprocess import preprocess_frame, preprocess_movement

        track_data = {}
        segments = track.get_segments(
            self.params.square_width**2,
            ffc_frames=[] if dont_filter else clip.ffc_frames,
            repeats=1,
            segment_frames=segment_frames,
            segment_types=self.params.segment_types,
            from_last=predict_from_last,
            max_segments=max_segments,
            dont_filter=dont_filter,
            filter_by_fp=False,
            min_segments=min_segments,
        )
        frame_indices = set()
        for segment in segments:
            frame_indices.update(set(segment.frame_indices))
        frame_indices = list(frame_indices)
        frame_indices.sort()

        # should really be over whole track buts let just do the indices we predict of
        #  seems to make little different to just doing a min max normalization
        thermal_norm_limits = None
        filtered_norm_limits = None
        if self.params.diff_norm or self.params.thermal_diff_norm:
            min_diff = None
            max_diff = 0
            thermal_max_diff = None
            thermal_min_diff = None
            for i, region in enumerate(reversed(track.bounds_history)):
                if region.blank:
                    continue
                if region.width == 0 or region.height == 0:
                    logging.warn(
                        "No width or height for frame %s regoin %s",
                        region.frame_number,
                        region,
                    )
                    continue
                f = clip.get_frame(region.frame_number)
                if region.blank or region.width <= 0 or region.height <= 0 or f is None:
                    continue

                f.float_arrays()

                if self.params.thermal_diff_norm:
                    diff_frame = f.thermal - np.median(f.thermal)
                    new_max = np.amax(diff_frame)
                    new_min = np.amin(diff_frame)
                    if thermal_min_diff is None or new_min < thermal_min_diff:
                        thermal_min_diff = new_min
                    if thermal_max_diff is None or new_max > thermal_max_diff:
                        thermal_max_diff = new_max
                if self.params.diff_norm:
                    diff_frame = region.subimage(f.filtered)
                    # - region.subimage(
                    #     clip.background
                    # )
                    new_max = np.amax(diff_frame)
                    new_min = np.amin(diff_frame)
                    if min_diff is None or new_min < min_diff:
                        min_diff = new_min
                    if new_max > max_diff:
                        max_diff = new_max
            if self.params.thermal_diff_norm:
                thermal_norm_limits = (thermal_min_diff, thermal_max_diff)

            if self.params.diff_norm:
                filtered_norm_limits = (min_diff, max_diff)
        for frame_index in frame_indices:
            region = track.bounds_history[frame_index - track.start_frame]

            frame = clip.get_frame(region.frame_number)
            # filtered is calculated slightly different for tracking, set to null so preprocess can recalc it
            if frame is None:
                logging.error(
                    "Clasifying clip %s track %s can't get frame %s",
                    clip.get_id(),
                    track.get_id(),
                    region.frame_number,
                )
                raise Exception(
                    "Clasifying clip {} track {} can't get frame {}".format(
                        clip.get_id(), track.get_id(), region.frame_number
                    )
                )
            cropped_frame = preprocess_frame(
                frame,
                (self.params.frame_size, self.params.frame_size),
                region,
                clip.background,
                clip.crop_rectangle,
                calculate_filtered=False,
                filtered_norm_limits=filtered_norm_limits,
                thermal_norm_limits=thermal_norm_limits,
            )
            track_data[frame.frame_number] = cropped_frame
        features = None
        if self.params.mvm:
            from ml_tools.forestmodel import process_track as forest_process_track

            features = forest_process_track(
                clip, track, normalize=True, predict_from_last=predict_from_last
            )

        preprocessed = []
        masses = []
        for segment in segments:
            segment_frames = []
            for frame_i in segment.frame_indices:
                f = track_data[frame_i]
                # probably no need to copy
                segment_frames.append(f.copy())
            frames = preprocess_movement(
                segment_frames,
                self.params.square_width,
                self.params.frame_size,
                self.params.channels,
                self.preprocess_fn,
                sample=f"{clip.get_id()}-{track.get_id()}",
            )
            if frames is None:
                logging.warn("No frames to predict on")
                continue
            preprocessed.append(frames)
            masses.append(segment.mass)
        preprocessed = np.array(preprocessed)
        if self.params.mvm:
            features = features[np.newaxis, :]
            features = np.repeat(features, len(preprocessed), axis=0)
            preprocessed = [preprocessed, features]

        return [s.frame_indices for s in segments], preprocessed, masses


class NeuralInterpreter(Interpreter):
    TYPE = "Neural"

    def __init__(self, model_name):
        from openvino.inference_engine import IENetwork, IECore

        super().__init__(model_name)
        model_name = Path(model_name)
        # can use to test on PC
        # device = "CPU"
        device = "MYRIAD"
        model_xml = model_name.with_suffix(".xml")
        model_bin = model_name.with_suffix(".bin")
        ie = IECore()
        ie.set_config({}, device)
        net = ie.read_network(model=model_xml, weights=model_bin)
        self.input_blob = next(iter(net.input_info))
        self.out_blob = next(iter(net.outputs))
        self.input_shape = net.input_info[self.input_blob].input_data.shape

        net.batch_size = 1
        self.exec_net = ie.load_network(network=net, device_name=device)
        self.preprocess_fn = inc3_preprocess

    def predict(self, input_x):
        if input_x is None:
            return None
        input_x = np.float32(input_x)
        channels_last = input_x.shape[-1] == 3
        if channels_last:
            input_x = np.moveaxis(input_x, 3, 1)
        res = self.exec_net.infer(inputs={self.input_blob: input_x})
        res = res[self.out_blob]
        return res

    def shape(self):
        return 1, self.input_shape


class LiteInterpreter(Interpreter):
    TYPE = "TFLite"

    def __init__(self, model_name):
        super().__init__(model_name)

        import tflite_runtime.interpreter as tflite

        model_name = Path(model_name)
        model_name = model_name.with_suffix(".tflite")
        self.interpreter = tflite.Interpreter(str(model_name))

        self.interpreter.allocate_tensors()  # Needed before execution!

        self.output = self.interpreter.get_output_details()[
            0
        ]  # Model has single output.
        self.input = self.interpreter.get_input_details()[0]  # Model has single input.
        self.preprocess_fn = self.get_preprocess_fn()
        # inc3_preprocess

    def predict(self, input_x):
        input_x = np.float32(input_x)
        preds = []
        # only works on input of 1
        for data in input_x:
            self.interpreter.set_tensor(self.input["index"], data[np.newaxis, :])
            self.interpreter.invoke()
            pred = self.interpreter.get_tensor(self.output["index"])
            preds.append(pred)
        return pred

    def shape(self):
        return 1, self.input["shape"]


def inc3_preprocess(x):
    x /= 127.5
    x -= 1.0
    return x


def get_interpreter(model):
    # model_name, type = os.path.splitext(model.model_file)

    logging.info(
        "Loading %s of type %s",
        model.model_file,
        model.type,
    )

    if model.type == LiteInterpreter.TYPE:
        classifier = LiteInterpreter(model.model_file)
    elif model.type == NeuralInterpreter.TYPE:
        classifier = NeuralInterpreter(model.model_file)
    elif model.type == "RandomForest":
        from ml_tools.forestmodel import ForestModel

        classifier = ForestModel(model.model_file)
    else:
        from ml_tools.kerasmodel import KerasModel

        classifier = KerasModel()
        classifier.load_model(model.model_file, weights=model.model_weights)
    classifier.id = model.id
    return classifier



import logging
from ml_tools.kerasmodel import KerasModel, grid_search
import pickle
import os
from ml_tools.logs import init_logging
from pathlib import Path
import absl.logging
import sys

# tensorflow stealing my log handler
root_logger = logging.getLogger()
for handler in root_logger.handlers:
    if isinstance(handler, logging.StreamHandler) and handler.stream == sys.stderr:
        root_logger.removeHandler(handler)


def remove_fp_segments(datasets, ignore_file):
    # testing removing some automatically selected bad clips
    unique_ids = ignore_clips(ignore_file)
    print("ignore ids", unique_ids)
    for dataset in datasets:
        delete_me = []
        for segment in dataset.segments:
            if segment.unique_track_id in unique_ids:
                dataset.segments_by_label[segment.label].remove(segment)
                del dataset.segments_by_id[segment.id]
                delete_me.append(segment)
                print("deleting segment", segment.unique_track_id)
        for delete in delete_me:
            try:
                dataset.remove_track(delete.track_id)
            except:
                pass
            dataset.segments.remove(delete)
            print("delete track", delete.track_id, " from", dataset.name)
        dataset.rebuild_cdf()


def train_model(
    run_name,
    conf,
    hyper_params,
    weights=None,
    do_grid_search=None,
    ignore=None,
    epochs=None,
    fine_tune=None,
):
    init_logging()
    """Trains a model with the given hyper parameters."""
    data_dir = Path(conf.base_folder) / "training-data"
    model = KerasModel(
        train_config=conf.train,
        labels=conf.labels,
        data_dir=data_dir,
    )

    logging.info("Importing datasets from %s ", data_dir)
    model.load_training_meta(data_dir)

    if do_grid_search:
        print("Searching hparams")
        grid_search(model)
        model.close()
        return
    print()
    print("Training started")
    print("---------------------")
    print("Hyper parameters")
    print("---------------------")
    print(model.hyperparams_string)
    print()
    print("EPOCHS ", epochs)
    try:
        model.train_model(
            epochs=conf.train.epochs if epochs is None else epochs,
            run_name=run_name,
            weights=weights,
            resample=False,
            rebalance=False,
            fine_tune=fine_tune,
        )
    except KeyboardInterrupt:
        pass
    except:
        logging.error("Exited with error ", exc_info=True)
    model.close()


def ignore_clips(file_path):
    ignore_clips = []
    with open(file_path) as stream:
        for line in stream:
            if line.strip() == "" or line[0] != "[":
                continue
            try:
                line = line.replace("[", "")
                line = line.replace("]", "")
                clips = line.split(",")

                for clip in clips:
                    clip = clip.replace("'", "")
                    ignore_clips.append(clip.strip())
            except:
                logging.warn(
                    "Could not parse clip_id %s from %s",
                    line,
                    file_path,
                )
    return ignore_clips



import time
import dbus
import logging
import json
from dbus.mainloop.glib import DBusGMainLoop

DBUS_NAME = "org.cacophony.Events"
DBUS_PATH = "/org/cacophony/Events"


def throttled_event():
    bus = dbus.SystemBus(mainloop=DBusGMainLoop())
    try:
        proxy = bus.get_object(DBUS_NAME, DBUS_PATH)
        proxy.Add("{}", "throttle", dbus.Int64(time.time_ns()))
    except:
        logging.error("throttle dbus error ", exc_info=True)


def log_event(event_type, details=None):
    print("TODO Fix event logging")
    return
    bus = dbus.SystemBus(mainloop=DBusGMainLoop())
    data = None
    if details is not None:
        data = {"description": {"details": details}}

    try:
        proxy = bus.get_object(DBUS_NAME, DBUS_PATH)
        if data is None:
            json_d = "{}"
        else:
            json_d = json.dumps(data)
        proxy.Add(json_d, event_type, dbus.Int64(time.time_ns()))
    except:
        logging.error("log event dbus error ", exc_info=True)



import math
import matplotlib.pyplot as plt

import tensorflow as tf
from functools import partial
import numpy as np
import time
from config.config import Config
import json
from ml_tools.logs import init_logging
import logging

from ml_tools.featurenorms import mean_v, std_v
from ml_tools.frame import TrackChannels
from pathlib import Path

# seed = 1341
# tf.random.set_seed(seed)
# np.random.seed(seed)
AUTOTUNE = tf.data.AUTOTUNE
# IMAGE_SIZE = [256, 256]
# BATCH_SIZE = 64

insect = None
fp = None


def get_excluded():
    return [
        "goat",
        "lizard",
        "not identifiable",
        "other",
        "pest",
        "pig",
        "sealion",
        "bat",
        "mammal",
        "frog",
        "cow",
        # added gp forretrain
        "wombat",
        "gray kangaroo",
        "echidna",
        "fox",
        "deer",
        "sheep",
        # "wombat",
    ]


def get_remapped(multi_label=False):
    land_bird = "land-bird" if multi_label else "bird"
    return {
        "echidna": "hedgehog",
        "grey kangaroo": "wallaby",
        "sambar deer": "deer",
        "mouse": "rodent",
        "rat": "rodent",
        "rain": "false-positive",
        "water": "false-positive",
        "insect": "false-positive",
        "allbirds": "bird",
        "black swan": land_bird,
        "brown quail": land_bird,
        "california quail": land_bird,
        "duck": land_bird,
        "pheasant": land_bird,
        "pukeko": land_bird,
        "quail": land_bird,
        "chicken": land_bird,
    }


def get_extra_mappings(labels):
    land_birds = ["land-bird"]
    if "bird" not in labels:
        return None
    bird_index = labels.index("bird")
    values = []
    keys = []
    for l in land_birds:
        if l in labels:
            l_i = labels.index(l)
            keys.append(l_i)
            values.append(bird_index)
    extra_label_map = tf.lookup.StaticHashTable(
        initializer=tf.lookup.KeyValueTensorInitializer(
            keys=tf.constant(keys),
            values=tf.constant(values),
        ),
        default_value=tf.constant(-1),
        name="extra_label_map",
    )
    logging.info("Extra label mapping is %s to %s ", keys, values)
    return extra_label_map


def load_dataset(filenames, remap_lookup, labels, args):
    deterministic = args.get("deterministic", False)
    ignore_order = tf.data.Options()
    ignore_order.experimental_deterministic = (
        deterministic  # disable order, increase speed
    )
    dataset = tf.data.TFRecordDataset(
        filenames, compression_type="GZIP", num_parallel_reads=4
    )
    dataset = dataset.with_options(
        ignore_order
    )  # uses data as soon as it streams in, rather than in its original order

    image_size = args["image_size"]
    labeled = args.get("labeled", True)
    augment = args.get("augment", False)
    preprocess_fn = args.get("preprocess_fn")
    include_features = args.get("include_features", False)
    only_features = args.get("only_features", False)
    one_hot = args.get("one_hot", True)
    dataset = dataset.apply(tf.data.experimental.ignore_errors())
    extra_label_map = None
    if args.get("multi_label"):
        extra_label_map = get_extra_mappings(labels)
        logging.info("Using multi label")
    dataset = dataset.map(
        partial(
            read_tfrecord,
            image_size=image_size,
            remap_lookup=remap_lookup,
            num_labels=len(labels),
            labeled=labeled,
            augment=augment,
            preprocess_fn=preprocess_fn,
            include_features=include_features,
            only_features=only_features,
            one_hot=one_hot,
            extra_label_map=extra_label_map,
            include_track=args.get("include_track", False),
            num_frames=args.get("num_frames", 25),
            channels=args.get(
                "channels", [TrackChannels.thermal.name, TrackChannels.filtered.name]
            ),
        ),
        num_parallel_calls=AUTOTUNE,
        deterministic=deterministic,
    )
    if only_features:
        filter_nan = lambda x, y: not tf.reduce_any(tf.math.is_nan(x))
    else:
        filter_nan = lambda x, y: not tf.reduce_any(tf.math.is_nan(x[0]))

    dataset = dataset.filter(filter_nan)

    # if features are missing they wil be 0 size
    if args.get("only_features"):
        filter_none = lambda x, y: tf.size(x) > 0
        dataset = dataset.filter(filter_none)
    elif args.get("include_features"):
        filter_none = lambda x, y: tf.size(x[1]) > 0
        dataset = dataset.filter(filter_none)
    return dataset


rotation_augmentation = tf.keras.Sequential(
    [
        tf.keras.layers.RandomRotation(0.1, fill_mode="nearest", fill_value=0),
    ]
)
data_augmentation = tf.keras.Sequential(
    [
        tf.keras.layers.RandomFlip("horizontal"),
        tf.keras.layers.RandomBrightness(0.2),  # better per frame or per sequence??
        tf.keras.layers.RandomContrast(0.5),
    ]
)


def read_tfrecord(
    example,
    image_size,
    remap_lookup,
    num_labels,
    labeled,
    augment=False,
    preprocess_fn=None,
    only_features=False,
    one_hot=True,
    include_features=False,
    extra_label_map=None,
    include_track=False,
    num_frames=25,
    channels=[TrackChannels.thermal.name, TrackChannels.filtered.name],
):
    logging.info(
        "Read tf record with image %s lbls %s labeld %s aug  %s  prepr %s only features %s one hot %s include fetures %s num frames %s",
        image_size,
        num_labels,
        labeled,
        augment,
        preprocess_fn,
        only_features,
        one_hot,
        include_features,
        num_frames,
    )
    load_images = not only_features
    tfrecord_format = {
        "image/class/label": tf.io.FixedLenFeature((), tf.int64, -1),
    }
    if load_images:
        if TrackChannels.filtered.name in channels:
            tfrecord_format["image/filteredencoded"] = tf.io.FixedLenSequenceFeature(
                [num_frames * 32 * 32], dtype=tf.float32, allow_missing=True
            )
        if TrackChannels.thermal.name in channels:
            tfrecord_format["image/thermalencoded"] = tf.io.FixedLenSequenceFeature(
                [num_frames * 32 * 32], dtype=tf.float32, allow_missing=True
            )

    if include_track:
        tfrecord_format["image/track_id"] = tf.io.FixedLenFeature((), tf.int64, -1)
        tfrecord_format["image/avg_mass"] = tf.io.FixedLenFeature((), tf.int64, -1)
    if include_features or only_features:
        tfrecord_format["image/features"] = tf.io.FixedLenSequenceFeature(
            [36 * 5 + 8], dtype=tf.float32, allow_missing=True
        )
    example = tf.io.parse_single_example(example, tfrecord_format)
    if load_images:
        if TrackChannels.thermal.name in channels:
            thermalencoded = example["image/thermalencoded"]
            thermals = tf.reshape(thermalencoded, [num_frames, 32, 32, 1])
        if TrackChannels.filtered.name in channels:
            filteredencoded = example["image/filteredencoded"]
            filtered = tf.reshape(filteredencoded, [num_frames, 32, 32, 1])
        rgb_image = None
        for type in channels:
            if type == TrackChannels.thermal.name:
                image = thermals
            elif type == TrackChannels.filtered.name:
                image = filtered
            if rgb_image is None:
                rgb_image = image
            else:
                rgb_image = tf.concat((rgb_image, image), axis=3)
        # rotation augmentation before tiling
        if augment:
            logging.info("Augmenting")
            rgb_image = rotation_augmentation(rgb_image)
        rgb_image = tf.ensure_shape(rgb_image, (num_frames, 32, 32, len(channels)))
        if num_frames > 1:
            rgb_image = tile_images(rgb_image)

        if augment:
            rgb_image = data_augmentation(rgb_image)
        if num_frames == 1:
            # remove the leading axis
            rgb_image = tf.squeeze(rgb_image)

        if preprocess_fn is not None:
            logging.info(
                "Preprocessing with %s.%s",
                preprocess_fn.__module__,
                preprocess_fn.__name__,
            )
            rgb_image = preprocess_fn(rgb_image)
    if labeled:
        label = tf.cast(example["image/class/label"], tf.int32)
        label = remap_lookup.lookup(label)
        if extra_label_map is not None:
            extra = extra_label_map.lookup(label)
            label = tf.stack([label, extra], axis=0)
        if one_hot:
            label = tf.one_hot(label, num_labels)
            if extra_label_map is not None:
                label = tf.reduce_max(label, axis=0)
        if include_track:

            track_id = tf.cast(example["image/track_id"], tf.int32)
            avg_mass = tf.cast(example["image/avg_mass"], tf.int32)
            label = (label, track_id, avg_mass)
        if include_features or only_features:
            features = tf.squeeze(example["image/features"])
            if only_features:
                return features, label
            return (rgb_image, features), label
        return rgb_image, label
    if only_features:
        return tf.squeeze(example["image/features"])
    elif include_features:
        return (rgb_image, tf.squeeze(example["image/features"]))
    return rgb_image


def tile_images(images):
    index = 0
    image = None
    for x in range(5):
        t_row = tf.concat(tf.unstack(images[index : index + 5]), axis=1)
        if image is None:
            image = t_row
        else:
            image = tf.concat([image, t_row], 0)

        index += 5
    return image


from collections import Counter


# test stuff
def main():
    init_logging()
    config = Config.load_from_file("classifier-thermal.yaml")
    from .tfdataset import get_dataset, get_distribution

    # file = "/home/gp/cacophony/classifier-data/thermal-training/cp-training/training-meta.json"
    training_folder = Path(config.base_folder) / "training-data"
    meta_f = training_folder / "training-meta.json"
    with open(meta_f, "r") as f:
        meta = json.load(f)
    labels = meta.get("labels", [])
    datasets = []

    resampled_ds, remapped, labels, epoch_size = get_dataset(
        # dir,
        load_dataset,
        training_folder / "test",
        labels,
        batch_size=32,
        image_size=(160, 160),
        # augment=True,
        # preprocess_fn=tf.keras.applications.inception_v3.preprocess_input,
        resample=False,
        include_features=False,
        remapped_labels=get_remapped(),
        excluded_labels=get_excluded(),
        include_track=True,
        num_frames=1,
    )
    print("Ecpoh size is", epoch_size)
    # print(get_distribution(resampled_ds, len(labels), extra_meta=False))
    # return
    #
    save_dir = Path("./test-images")
    save_dir.mkdir(exist_ok=True)
    for e in range(1):
        batch_i = 0
        print("epoch", e)
        for x, y in resampled_ds:
            show_batch(x, y, labels, save=save_dir / f"{batch_i}.jpg", tracks=True)
            batch_i += 1
    # return


def show_batch(image_batch, label_batch, labels, save=None, tracks=False):
    plt.figure(figsize=(10, 10))
    print("images in batch", len(image_batch), len(label_batch))
    num_images = min(len(image_batch), 25)
    if tracks:
        track_batch = label_batch[1]
        label_batch = label_batch[0]
    for n in range(num_images):
        ax = plt.subplot(5, 5, n + 1)
        img = np.uint8(image_batch[n])
        channels = img.shape[-1]
        repeat = 3 - channels
        while repeat > 0:
            img = np.concatenate((img, img[:, :, :1]), axis=2)
            repeat -= 1
        # if repeat > 0:
        # print(img.shape, " repeating", repeat)
        plt.imshow(img)
        if tracks:
            plt.title(f"{labels[np.argmax(label_batch[n])]}-{track_batch[n]}")
        else:
            plt.title(labels[np.argmax(label_batch[n])])

        plt.axis("off")
    # return
    if save:
        plt.savefig(save)
    plt.show()


if __name__ == "__main__":
    main()



import sys
import logging


def init_logging(timestamps=False):
    """Set up logging for use by various classifier pipeline scripts.

    Logs will go to stderr.
    """
    root_logger = logging.getLogger()
    for handler in root_logger.handlers:
        if isinstance(handler, logging.StreamHandler) and handler.stream == sys.stderr:
            root_logger.removeHandler(handler)
    fmt = "%(process)d %(thread)s:%(levelname)7s %(message)s"
    if timestamps:
        fmt = "%(asctime)s " + fmt
    logging.basicConfig(
        stream=sys.stderr, level=logging.INFO, format=fmt, datefmt="%Y-%m-%d %H:%M:%S"
    )



# Build a segment dataset for training.
# Segment headers will be extracted from a track database and balanced
# according to class. Some filtering occurs at this stage as well, for example
# tracks with low confidence are excluded.

import argparse
import os
import random
import datetime
import logging
import pickle
import pytz
import json
from dateutil.parser import parse as parse_date
from ml_tools.logs import init_logging
from config.config import Config
from ml_tools.dataset import Dataset
from ml_tools.datasetstructures import Camera
from ml_tools.tfwriter import create_tf_records
from ml_tools.irwriter import save_data as save_ir_data
from ml_tools.thermalwriter import save_data as save_thermal_data
from ml_tools.tools import CustomJSONEncoder
import attrs
import numpy as np

from pathlib import Path

MAX_TEST_TRACKS = None
MAX_TEST_SAMPLES = None

MIN_SAMPLES = 1
MIN_TRACKS = 1
# LOW_SAMPLES_LABELS = ["bird", "cat", "possum"]
LOW_SAMPLES_LABELS = []


def load_config(config_file):
    return Config.load_from_file(config_file)


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-m",
        "--min-samples",
        default=None,
        type=int,
        help="Min tracks per dataset (Default 100)",
    )
    parser.add_argument(
        "-a",
        "--aug_percent",
        default=None,
        type=float,
        help="Percentage of training set to add extra augmentations of",
    )
    parser.add_argument("--split-file", help="Json file defining a split")
    parser.add_argument(
        "--ext", default=".cptv", help="Extension of files to load .mp4,.cptv,.hdf5"
    )

    parser.add_argument("-c", "--config-file", help="Path to config file to use")
    parser.add_argument("-d", "--date", help="Use clips after this")
    parser.add_argument("--dont-cap", action="count", help="Dont cap numbers")
    parser.add_argument(
        "--consecutive-segments",
        action="count",
        default=False,
        help="Use consecutive frames for segments",
    )
    parser.add_argument("data_dir", help="Directory of hdf5 files")
    args = parser.parse_args()
    if args.date:
        # if args.date == "None":
        #     args.date = None
        # else:
        args.date = parse_date(args.date)
        args.date = args.date.replace(tzinfo=pytz.utc)
    # else:
    #     if args.date is None:
    #         args.date = datetime.datetime.now(pytz.utc) - datetime.timedelta(days=30)

    if args.min_samples is not None:
        global MAX_TEST_TRACKS, MAX_TEST_SAMPLES, MIN_SAMPLES, MIN_TRACKS
        MAX_TEST_TRACKS = args.min_samples
        MAX_TEST_SAMPLES = args.min_samples

        MIN_SAMPLES = args.min_samples
        MIN_TRACKS = args.min_samples
        # args.date = datetime.datetime.now() - datetime.timedelta(days=30)
    if args.data_dir is not None:
        args.data_dir = Path(args.data_dir)
    logging.info("Loading training set up to %s", args.date)
    return args


def show_clips_breakdown(dataset):
    print("Clips breakdown:")

    for label in dataset.labels:
        print(label)
        lbl_samples = dataset.samples_by_label.get(label, [])
        if dataset.label_mapping:
            actual_labels = {}
            for s in lbl_samples:
                actual_labels.setdefault(s.original_label, set()).add(s.clip_id)
            for actual, clips in actual_labels.items():
                print("  {:<20} {} clips".format(actual, len(clips)))

        else:
            clips = [s.clip_id for s in lbl_samples]
            clips = set(clips)
            print("  {:<20} {} clips".format(label, len(clips)))


def show_tracks_breakdown(dataset):
    print("Tracks breakdown:")

    for label in dataset.labels:
        lbl_samples = dataset.samples_by_label.get(label, [])
        tracks = [s.unique_track_id for s in lbl_samples]
        tracks = set(tracks)
        print("  {:<20} {} tracks".format(label, len(tracks)))


def show_cameras_tracks(dataset):
    for id, camera in dataset.cameras_by_id.items():
        count = "Tracks:"
        for label in dataset.labels:
            count = "{} {}: {}".format(count, label, camera.label_track_count(label))
        print("Camera", id, count)


def show_cameras_breakdown(dataset):
    print("Cameras breakdown")
    samples_by_camera = {}
    for clip in dataset.clips:
        if clip.camera not in samples_by_camera:
            samples_by_camera[clip.camera] = []

        for track in clip.tracks:
            samples_by_camera[clip.camera].extend(track.samples)

    for camera, samples in samples_by_camera.items():
        print(
            "Camera: {:<20} {}".format(
                "None" if camera is None else camera, len(samples)
            )
        )


def show_bins_breakdown(dataset):
    print("Bins breakdown")
    samples_by_bins = {}
    for clip in dataset.clips:
        for t in clip.tracks:
            for sample in t.samples:
                if sample.bin_id not in samples_by_bins:
                    samples_by_bins[sample.bin_id] = []
                samples_by_bins[sample.bin_id].append(sample)

    for bin_id, samples in samples_by_bins.items():
        print(
            "Bin Id: {:<20} {}".format(
                "None" if bin_id is None else bin_id, len(samples)
            )
        )


def show_stations_breakdown(dataset):
    print("Stations breakdown")
    samples_by_station = {}
    for clip in dataset.clips:
        if clip.station_id not in samples_by_station:
            samples_by_station[clip.station_id] = []

        for track in clip.tracks:
            samples_by_station[clip.station_id].extend(track.samples)

    for station, samples in samples_by_station.items():
        print(
            "StationId: {:<20} {}".format(
                "None" if station is None else station, len(samples)
            )
        )


def show_samples_breakdown(dataset):
    print("Samples breakdown:")
    for label in dataset.labels:
        count = len(dataset.samples_by_label.get(label, []))
        print("  {:<20} {} Samples".format(label, count))


def print_cameras(train, validation, test):
    print("Cameras per set:")
    print("-" * 90)
    print("Train")
    print(train.camera_names)
    print("Validation")
    print(validation.camera_names)
    print("Test")
    print(test.camera_names)

    print()


def print_counts(train, validation, test):
    logging.info("Counts per class:")
    logging.info("-" * 90)
    logging.info(
        "{:<20} {:<21} {:<21} {:<21}".format("Class", "Train", "Validation", "Test")
    )
    logging.info("-" * 90)
    logging.info("Samples / Tracks/ Bins/ weight")
    # display the dataset summary
    for label in train.labels:
        logging.info(
            "{:<20} {:<20} {:<20} {:<20}".format(
                label,
                "{}/{}/{}/{:.1f}".format(*train.get_counts(label)),
                "{}/{}/{}/{:.1f}".format(*validation.get_counts(label)),
                "{}/{}/{}/{:.1f}".format(*test.get_counts(label)),
            )
        )
    logging.info("")


# default split is by stationid, but some labels dont have many stations so best to just split by clip
split_by_clip = ["vehicle", "penguin", "wallaby"]


def split_label(
    dataset,
    label,
    counts,
    train_count,
    validation_count,
    test_count,
    existing_test_count=0,
    max_samples=None,
    use_test=True,
):
    # split a label from dataset such that vlaidation is 15% or MIN_TRACKS
    # dont split these by location and camera
    logging.info(
        "Splitting %s have counts (Tracks/Samples/Bins) %s already have (Tracks/Samples) in train %s validation %s and in test %s",
        label,
        counts,
        train_count,
        validation_count,
        test_count,
    )

    samples = dataset.samples_by_label.get(label, [])
    sample_bins = set([sample.bin_id for sample in samples])

    if counts[2] < 4 or counts[0] < 100:
        global split_by_clip
        split_by_clip.append(label)

    if label in split_by_clip:
        sample_bins = set([sample.clip_id for sample in samples])
        logging.info("%s Splitting by clip", label)
        samples_by_bin = {}
        for clip in dataset.clips:
            if clip.clip_id in sample_bins:
                samples = clip.get_samples()
                by_id = {}
                for s in samples:
                    by_id[s.id] = s
                samples_by_bin[clip.clip_id] = by_id
    else:
        samples_by_bin = dataset.samples_by_bin
    if len(sample_bins) == 0:
        return None, None, None

    if max_samples is not None:
        sample_bins = np.random.choice(
            sample_bins, min(len(sample_bins), max_samples), replace=False
        )

    sample_count = counts[1]
    total_tracks = counts[0]

    sample_bins = list(sample_bins)

    random.shuffle(sample_bins)
    train_c = []
    validate_c = []
    test_c = [] if use_test else None

    camera_type = "validate"
    add_to = validate_c
    last_index = 0
    label_count = 0
    min_t = MIN_SAMPLES

    if label in LOW_SAMPLES_LABELS:
        min_t = 10
    num_validate_samples = max(sample_count * 0.15, min_t) - validation_count[1]
    num_test_samples = max(sample_count * 0.05, min_t)
    if MAX_TEST_SAMPLES is not None:
        num_test_samples = min(MAX_TEST_SAMPLES, num_test_samples)
    num_test_samples -= test_count[1]

    min_t = MIN_TRACKS

    if label in LOW_SAMPLES_LABELS:
        min_t = 10

    num_validate_tracks = max(total_tracks * 0.15, min_t) - validation_count[0]
    num_test_tracks = max(total_tracks * 0.05, min_t)
    if MAX_TEST_TRACKS is not None:
        num_test_tracks = min(MAX_TEST_TRACKS, num_test_tracks)
    num_test_tracks -= test_count[0]

    track_limit = num_validate_tracks
    sample_limit = num_validate_samples
    tracks = set()
    logging.info(
        f"{label} - looking for val {num_validate_tracks} tracks out of {total_tracks} tracks and {num_validate_samples} samples from a total of {sample_count} samples  with {num_test_tracks} test tracks and {num_test_samples} test samples"
    )
    # if sample_limit < 0 and track_limit < 0:
    #     add_to = test_c
    #     camera_type = "test"
    #     sample_limit = num_test_samples
    #     track_limit = num_test_tracks
    #     label_count = 0
    #     tracks = set()
    if sample_limit > 0 and track_limit > 0:
        for i, sample_bin in enumerate(sample_bins):
            samples = samples_by_bin[sample_bin].values()
            for sample in samples:
                if sample.label == label:
                    tracks.add(sample.track_id)
                    label_count += 1

                # sample.camera = "{}-{}".format(sample.camera, camera_type)
                add_to.append(sample)

                if label in split_by_clip:
                    dataset.remove_sample(sample)
            if label not in split_by_clip:
                # while len(samples) > 0:
                sample_ids = [(s.id, s.bin_id) for s in samples]
                for id, bin_id in sample_ids:
                    dataset.remove_sample_by_id(id, bin_id)
            samples_by_bin[sample_bin] = {}
            last_index = i
            track_count = len(tracks)
            if label_count >= sample_limit and track_count >= track_limit:
                # 100 more for test
                if add_to == validate_c:
                    add_to = test_c
                    camera_type = "test"
                    if num_test_samples <= 0:
                        break
                    sample_limit = num_test_samples
                    track_limit = num_test_tracks
                    label_count = 0
                    tracks = set()
                    if use_test is False:
                        break
                else:
                    break

    # sample_bins = sample_bins[last_index + 1 :]
    # clearining anyway
    camera_type = "train"
    added = 0
    for i, sample_bin in enumerate(sample_bins):
        samples = samples_by_bin[sample_bin].values()
        for sample in samples:
            # sample.camera = "{}-{}".format(sample.camera, camera_type)
            train_c.append(sample)
            added += 1

            if label in split_by_clip:
                dataset.remove_sample(sample)
        if label not in split_by_clip:
            sample_ids = [(s.id, s.bin_id) for s in samples]
            for id, bin_id in sample_ids:
                dataset.remove_sample_by_id(id, bin_id)

        samples_by_bin[sample_bin] = []
    return train_c, validate_c, test_c


def get_test_set_camera(dataset, test_clips, after_date):
    # load test set camera from tst_clip ids and all clips after a date
    print("after date is", after_date)
    test_samples = []
    test_clips = [
        clip
        for clip in dataset.clips
        if clip.clip_id in test_clips
        or after_date is not None
        and clip.rec_time.replace(tzinfo=pytz.utc) > after_date
    ]
    for clip in test_clips:
        for track in clip.tracks:
            for sample in track.samples:
                dataset.remove_sample(sample)
                test_samples.append(sample)

    return test_samples


def split_by_file(dataset, config, split_file, base_dir, make_val=True):
    base_dir = Path(base_dir)
    with open(split_file, "r") as f:
        split = json.load(f)

    samples_by_source = {}
    for s in dataset.samples_by_id.values():
        samples_by_source.setdefault(s.source_file.name, []).append(s)
    datasets = []
    for name in ["train", "validation", "test"]:
        logging.info("Loading %s", name)
        split_dataset = Dataset(
            dataset.dataset_dir,
            name,
            config,
            label_mapping=dataset.label_mapping,
            ext=dataset.ext,
            raw=dataset.raw,
        )
        if name == "train":
            split_dataset.enable_augmentation = True
        elif name == "test":
            split_dataset.skip_ffc = False
        split_files = split.get(name, [])
        for f in split_files:
            # print("loading", base_dir / f["source"])
            # continue
            source_file = base_dir / f["source"]
            if source_file.exists():
                try:
                    # can filter crappy segments here, or can do at train stage to have some way of testing different thresholds
                    split_dataset.load_clip(source_file, dont_filter_segment=True)
                except:
                    logging.error("Could not load %s", source_file, exc_info=True)
            else:
                pass
                # logging.warn("No source file %s found for %s", f, name)
        datasets.append(split_dataset)

    print_counts(*datasets)
    if make_val:
        train, val, _ = split_randomly(datasets[0], config, None, use_test=False)
        datasets = [train, val, datasets[2]]
    return datasets


def split_randomly(dataset, config, date, test_clips=[], use_test=True):
    # split data randomly such that a clip is only in one dataset
    # have tried many ways to split i.e. location and cameras found this is simplest
    # and the results are the same
    train = Dataset(
        dataset.dataset_dir,
        "train",
        config,
        label_mapping=dataset.label_mapping,
        raw=dataset.raw,
        ext=dataset.ext,
    )
    train.enable_augmentation = True
    validation = Dataset(
        dataset.dataset_dir,
        "validation",
        config,
        label_mapping=dataset.label_mapping,
        raw=dataset.raw,
        ext=dataset.ext,
    )
    test = None
    test_counts = {}

    if use_test:
        test_c = get_test_set_camera(dataset, test_clips, date)
        test = Dataset(
            dataset.dataset_dir,
            "test",
            config,
            label_mapping=dataset.label_mapping,
            raw=dataset.raw,
            ext=dataset.ext,
        )
        add_samples(dataset.labels, test, test_c, test_counts)

    validate_cameras = []
    train_cameras = []
    min_label = None
    for label in dataset.labels:
        label_count = len(dataset.samples_by_label.get(label, []))
        if label not in ["insect", "false-positive"]:
            continue
        if min_label is None or label_count < min_label[1]:
            min_label = (label, label_count)
    lbl_order = sorted(
        dataset.labels,
        key=lambda lbl: len(dataset.samples_by_label.get(lbl, [])),
    )
    if "wallaby" in lbl_order:
        # make sure we do wallaby first so we get a good split
        lbl_order.remove("wallaby")
        lbl_order.insert(0, "wallaby")
    if "pest" in lbl_order:
        # make sure its last
        lbl_order.remove("pest")
        lbl_order.append("pest")

    lbl_counts = {}
    for lbl in dataset.labels:
        samples = dataset.samples_by_label.get(lbl, [])
        tracks = set([s.track_id for s in samples])
        bins = set([s.bin_id for s in samples])

        lbl_counts[lbl] = (len(tracks), len(samples), len(bins))

    logging.debug("lbl order is %s", lbl_order)
    train_counts = {}
    validation_counts = {}
    existing_test_count = 0
    for label in lbl_order:
        # existing_test_count = len(test.samples_by_label.get(label, []))
        train_c, validate_c, test_c = split_label(
            dataset,
            label,
            counts=lbl_counts[label],
            train_count=train_counts.get(label, (0, 0)),
            validation_count=validation_counts.get(label, (0, 0)),
            test_count=test_counts.get(label, (0, 0)),
            existing_test_count=existing_test_count,
            use_test=use_test,
            # max_samples=min_label[1],
        )
        if train_c is not None:
            add_samples(dataset.labels, train, train_c, train_counts)
        if validate_c is not None:
            add_samples(dataset.labels, validation, validate_c, validation_counts)
        if test_c is not None:
            add_samples(dataset.labels, test, test_c, test_counts)
        logging.debug("Train counts %s", train_counts)
        logging.debug("val counts %s", validation_counts)

    return train, validation, test


def add_samples(
    labels,
    dataset,
    samples,
    counts,
):
    by_labels = {}
    counts = {}
    for s in samples:
        if s.label not in by_labels:
            by_labels[s.label] = []
        by_labels[s.label].append(s)

    for label, lbl_samples in by_labels.items():
        track_count = len(set([s.track_id for s in lbl_samples]))
        counts[label] = (
            track_count,
            len(lbl_samples),
        )
    dataset.add_samples(samples)


def validate_datasets(datasets, test_bins, after_date):
    # check that clips are only in one dataset
    # that only test set has clips after date
    # that test set is the only dataset with test_clips

    # for dataset in datasets[:2]:
    #     for track in dataset.tracks:
    #         assert track.start_time < date

    for i, dataset in enumerate(datasets[:2]):
        dont_check = set(
            [
                sample.bin_id
                for sample in dataset.samples_by_id.values()
                if sample.label in split_by_clip
            ]
        )
        bins = set([sample.bin_id for sample in dataset.samples_by_id.values()])
        clips = set([sample.clip_id for sample in dataset.samples_by_id.values()])

        bins = bins - dont_check
        if test_bins is not None and dataset.name != "test":
            assert (
                len(bins.intersection(set(test_bins))) == 0
            ), "test bins should only be in test set"
        if len(bins) == 0:
            continue
        for other in datasets[(i + 1) :]:
            if dataset.name == other.name:
                continue
            dont_check = set(
                [
                    sample.bin_id
                    for sample in other.samples_by_id.values()
                    if sample.label in split_by_clip
                ]
            )
            if other.name == "test" and after_date is not None:
                dont_check_other = set(
                    [
                        sample.bin_id
                        for sample in other.samples_by_id.values()
                        if sample.rec_time > after_date
                    ]
                )
                dont_check = dont_check | dont_check_other
            other_bins = set([sample.bin_id for sample in other.samples_by_id.values()])
            other_bins = other_bins - dont_check
            other_clips = set(
                [sample.clip_id for sample in other.samples_by_id.values()]
            )

            intersection = bins.intersection(set(other_bins))

            assert (
                len(bins.intersection(set(other_bins))) == 0
            ), "bins should only be in one set"

            assert (
                len(clips.intersection(set(other_clips))) == 0
            ), "clips should only be in one set"


land_birds = [
    "pukeko",
    "california quail",
    "brown quail",
    "black swan",
    "quail",
    "pheasant",
    "penguin",
    "duck",
    "chicken",
    "rooster",
]

label_paths_dl = "https://raw.githubusercontent.com/TheCacophonyProject/cacophony-web/main/api/classifications/label_paths.json"


def dl_mappings():
    import requests

    logging.info("Downloading mappings file from %s ", label_paths_dl)
    response = requests.get(label_paths_dl)
    response.raise_for_status()

    mapping_content = response.content.decode()
    print(mapping_content)
    with open("label_paths.json", "w") as f:
        f.write(mapping_content)
    return mapping_content


def get_mappings():
    labels_path = Path("label_paths.json")
    if not labels_path.exists():
        label_paths = json.loads(dl_mappings())
    with open("label_paths.json", "r") as f:
        label_paths = json.load(f)
    regroup = {}
    for l, path in label_paths.items():
        if l in land_birds:
            regroup[l] = l
            continue
        split_path = path.split(".")
        if len(split_path) == 1:
            regroup[l] = l
        elif path.startswith("all.mammal"):
            if len(split_path) == 4:
                regroup[l] = split_path[-2]
            else:
                regroup[l] = l
        else:
            # print("l", l, " has ", path)
            parent = split_path[-2]
            # print("Parent is", parent, path)
            if parent == "kiwi" or split_path[-1] == "kiwi":
                regroup[l] = "kiwi"
            elif parent == "other":
                regroup[l] = l

            else:
                if "bird." in path:
                    regroup[l] = "bird"

                elif len(split_path) > 2:
                    regroup[l] = split_path[-3]
                else:
                    regroup[l] = split_path[-1]
    return regroup


def dump_split_ids(datasets, out_file="datasplit.json"):
    splits = {}
    logging.info("Wrinting split ids to %s", out_file)
    for d in datasets:
        samples_by_source = d.get_samples_by_source()
        clips = []
        for source, samples in samples_by_source.items():
            tags = set([s.label for s in samples])
            clips.append(
                {
                    "clip_id": samples[0].clip_id,
                    "source": str(source),
                    "station_id": "{}".format(samples[0].station_id),
                    "tags": list(tags),
                }
            )
        # clips = set(clips)
        splits[d.name] = clips
    with open(out_file, "w") as f:
        json.dump(splits, f)
    return


def rough_balance(datasets):
    dev_threshold = 2000
    logging.info("Roughly Balancing")
    print_counts(*datasets)

    for dataset in datasets:
        lbl_counts = {}
        counts = []
        for label in dataset.labels:
            label_count = len(dataset.samples_by_label.get(label, []))
            lbl_counts[label] = label_count
            counts.append(label_count)
        counts.sort()
        std_dev = np.std(counts)
        logging.info("Counts are %s std dev %s", counts, std_dev)
        if std_dev < dev_threshold or len(counts) <= 1:
            logging.info("Not balancing")
            continue
        if len(counts) <= 2:
            cap_at = counts[-2]
        elif len(counts) < 7:
            cap_at = counts[-2]
        else:
            cap_at = counts[-2]
        logging.info("Capping dataset %s at %s", dataset.name, cap_at)
        for lbl, count in lbl_counts.items():
            if count <= cap_at:
                continue
            samples_to_remove = count - cap_at
            by_labels = dataset.samples_by_label[lbl]
            np.random.shuffle(by_labels)
            for i in range(samples_to_remove):
                dataset.remove_sample(by_labels[i])
    print_counts(*datasets)


def main():
    init_logging()
    args = parse_args()
    config = load_config(args.config_file)
    logging.info("Building for type %s", config.train.type)

    test_clips = config.build.test_clips()
    if test_clips is None:
        test_clips = []
    logging.info("# of test clips are %s", len(test_clips))
    label_mapping = get_mappings()
    logging.info("Using mappings %s", label_mapping)
    master_dataset = Dataset(
        args.data_dir,
        "dataset",
        config,
        consecutive_segments=args.consecutive_segments,
        label_mapping=label_mapping,
        raw=False if args.ext == ".hdf5" else True,
        ext=args.ext,
    )
    base_dir = Path(config.base_folder)
    record_dir = base_dir / "training-data"
    record_dir.mkdir(parents=True, exist_ok=True)

    if args.split_file:
        logging.info("Loading datasets from split file %s", args.split_file)
        datasets = split_by_file(master_dataset, config, args.split_file, args.data_dir)
        labels = set()
        for dataset in datasets:
            labels.update(dataset.labels)
        labels = list(labels)
        labels.sort()
        logging.info("Setting labels to %s", dataset.labels)
        for dataset in datasets:
            dataset.labels = labels
    else:
        tracks_loaded, total_tracks = master_dataset.load_clips(
            dont_filter_segment=True
        )

        master_dataset.labels.sort()
        print(
            "Loaded {}/{} tracks, found {:.1f}k samples".format(
                tracks_loaded, total_tracks, len(master_dataset.clips) / 1000
            )
        )
        for key, value in master_dataset.filtered_stats.items():
            if value != 0:
                print("  {} filtered {}".format(key, value))

        print()
        show_clips_breakdown(master_dataset)
        print()
        show_samples_breakdown(master_dataset)
        print()
        show_cameras_breakdown(master_dataset)
        print()
        show_stations_breakdown(master_dataset)
        print()
        show_bins_breakdown(master_dataset)
        print()
        print("Splitting data set into train / validation")

        datasets = split_randomly(master_dataset, config, args.date, test_clips)

        rough_balance(datasets)
        validate_datasets(datasets, test_clips, args.date)
        dump_split_ids(datasets, record_dir / "datasplit.json")

    print_counts(*datasets)
    print("split data")

    dataset_counts = {}
    # create_tf_records = create_thermal_records
    if config.train.type == "IR":
        threshold = (
            config.tracking[config.train.type]
            .motion.threshold_for_model(config.train.type)
            .background_thresh
        )
        # create_tf_records = create_ir_records
    else:
        threshold = None

    # create pre augmented records if samples are very low
    # not in use probably needs to be re tested if used
    aug_percent = args.aug_percent
    if aug_percent is not None:
        train_set = datasets[0]
        for l, samples in train_set.samples_by_label.items():
            track_dic = {}
            for s in samples:
                track_dic.setdefault(s.track_id, []).append(s)
            # track_dic = dict((x.track_id, x) for x in samples)

            for track, s in track_dic.items():
                augment_samples = int(aug_percent * len(s))
                # words = ['banana', 'pie', 'Washington', 'book']
                samples_by_mass = sorted(s, key=lambda s: s.region.mass, reverse=True)
                samples = np.random.choice(
                    samples_by_mass, augment_samples, replace=False
                )
                new_samples = []
                for s in samples:
                    new = s.copy()
                    new.augment = True
                    new_samples.append(new)
                train_set.add_samples(new_samples)
        print("Count post augmentation")
        print_counts(*datasets)

    for dataset in datasets:
        dir = os.path.join(record_dir, dataset.name)
        extra_args = {
            "use_segments": master_dataset.use_segments,
            "label_mapping": dataset.label_mapping,
        }
        if config.train.type == "IR":
            extra_args["back_thresh"] = threshold
            create_tf_records(
                dataset,
                dir,
                datasets[0].labels,
                save_ir_data,
                num_shards=100,
                back_thresh=threshold,
            )
        else:
            if args.ext != ".hdf5":
                extra_args.update(
                    {
                        "segment_frame_spacing": master_dataset.segment_spacing * 9,
                        "segment_width": master_dataset.segment_length,
                        "segment_types": master_dataset.segment_types,
                        "segment_min_avg_mass": master_dataset.segment_min_avg_mass,
                        "max_segments": master_dataset.max_segments,
                        "dont_filter_segment": True,
                        "skip_ffc": True,
                        "tag_precedence": config.build.tag_precedence,
                        "min_mass": master_dataset.min_frame_mass,
                        "thermal_diff_norm": config.build.thermal_diff_norm,
                        "filter_by_lq": master_dataset.filter_by_lq,
                        "max_frames": master_dataset.max_frames,
                    }
                )
            # dont filter the test set,
            extra_args["filter_by_fp"] = dataset.name != "test"
            create_tf_records(
                dataset,
                dir,
                datasets[0].labels,
                save_thermal_data,
                num_shards=100,
                num_frames=dataset.segment_length,
                **extra_args,
            )
        counts = {}
        for label in dataset.labels:
            count = len(dataset.samples_by_label.get(label, []))
            counts[label] = count
        dataset_counts[dataset.name] = counts
    # dont need dataset anymore just need some meta
    meta_filename = f"{record_dir}/training-meta.json"
    meta_data = {
        "labels": datasets[0].labels,
        "type": config.train.type,
        "counts": dataset_counts,
        "by_label": False,
        "config": attrs.asdict(config),
        "segment_types": master_dataset.segment_types,
    }

    with open(meta_filename, "w") as f:
        json.dump(meta_data, f, indent=4, cls=CustomJSONEncoder)


if __name__ == "__main__":
    main()



import gc
import json
import logging
import os.path
import time

import numpy as np

import cv2
from classify.trackprediction import Predictions
from track.clip import Clip
from track.cliptrackextractor import ClipTrackExtractor, is_affected_by_ffc
from ml_tools import tools
from track.irtrackextractor import IRTrackExtractor
from ml_tools.previewer import Previewer
from ml_tools.interpreter import get_interpreter
from track.trackextractor import extract_file
from classify.thumbnail import get_thumbnail_info, best_trackless_thumb


class ClipClassifier:
    """Classifies tracks within CPTV files."""

    # skips every nth frame.  Speeds things up a little, but reduces prediction quality.
    FRAME_SKIP = 1

    def __init__(self, config, model=None):
        """Create an instance of a clip classifier"""

        self.config = config
        # super(ClipClassifier, self).__init__(config, tracking_config)
        self.model = model
        # prediction record for each track

        self.previewer = Previewer.create_if_required(config, config.classify.preview)

        self.models = {}

    def load_models(self):
        for model in self.config.classify.models:
            logging.info("Loading %s", model)
            classifier = self.get_classifier(model)

    def get_classifier(self, model):
        """
        Returns a classifier object, which is created on demand.
        This means if the ClipClassifier is copied to a new process a new Classifier instance will be created.
        """
        if model.id in self.models:
            return self.models[model.id]
        load_start = time.time()
        logging.info("classifier loading %s", model.model_file)
        classifier = get_interpreter(model)
        logging.info("classifier loaded (%s)", time.time() - load_start)
        self.models[model.id] = classifier
        return classifier

    def get_meta_data(self, filename):
        """Reads meta-data for a given cptv file."""
        source_meta_filename = os.path.splitext(filename)[0] + ".txt"
        if os.path.exists(source_meta_filename):
            meta_data = tools.load_clip_metadata(source_meta_filename)

            tags = set()
            for record in meta_data["Tags"]:
                # skip automatic tags
                if record.get("automatic", False):
                    continue
                else:
                    tags.add(record["animal"])

            tags = list(tags)

            if len(tags) == 0:
                tag = "no tag"
            elif len(tags) == 1:
                tag = tags[0] if tags[0] else "none"
            else:
                tag = "multi"
            meta_data["primary_tag"] = tag
            return meta_data
        else:
            return None

    def process(
        self,
        source,
        cache=None,
        reuse_frames=None,
        track=False,
        calculate_thumbnails=False,
    ):
        # IF passed a dir extract all cptv files, if a cptv just extract this cptv file
        if not os.path.exists(source):
            logging.error("Could not find file or directory %s", source)
            return
        if os.path.isfile(source):
            self.process_file(
                source,
                cache=cache,
                reuse_frames=reuse_frames,
                track=track,
                calculate_thumbnails=calculate_thumbnails,
            )
            return
        for folder_path, _, files in os.walk(source):
            for name in files:
                if os.path.splitext(name)[1] in [".mp4", ".cptv", ".avi"]:
                    full_path = os.path.join(folder_path, name)
                    self.process_file(
                        full_path,
                        cache=cache,
                        reuse_frames=reuse_frames,
                        track=track,
                        calculate_thumbnails=calculate_thumbnails,
                    )

    def process_file(
        self,
        filename,
        cache=None,
        reuse_frames=None,
        track=False,
        calculate_thumbnails=False,
    ):
        """
        Process a file extracting tracks and identifying them.
        :param filename: filename to process
        :param enable_preview: if true an MPEG preview file is created.
        """
        _, ext = os.path.splitext(filename)
        cache_to_disk = (
            cache if cache is not None else self.config.classify.cache_to_disk
        )

        if track:
            logging.info("Doing tracking")
            clip, track_extractor = extract_file(
                filename, self.config, cache_to_disk, to_stdout=False
            )
        elif ext == ".cptv":
            track_extractor = ClipTrackExtractor(
                self.config.tracking,
                self.config.use_opt_flow,
                cache_to_disk,
                do_tracking=track,
                calculate_filtered=True,
                verbose=self.config.verbose,
                calculate_thumbnail_info=calculate_thumbnails,
            )
            logging.info("Using clip extractor")

        elif ext in [".avi", ".mp4"]:
            track_extractor = IRTrackExtractor(self.config.tracking, cache_to_disk)
            logging.info("Using ir extractor")
        else:
            logging.error("Unknown extention %s", ext)
            return False

        base_filename = os.path.splitext(os.path.basename(filename))[0]
        meta_file = os.path.join(os.path.dirname(filename), base_filename + ".txt")
        if not os.path.exists(filename):
            logging.error("File %s not found.", filename)
            return False
        if not os.path.exists(meta_file):
            logging.error("File %s not found.", meta_file)
            return False
        meta_data = tools.load_clip_metadata(meta_file)

        logging.info("Processing file '{}'".format(filename))

        if not track:
            clip = Clip(track_extractor.config, filename)
            clip.load_metadata(
                meta_data,
                self.config.build.tag_precedence,
            )
            track_extractor.parse_clip(clip)

        predictions_per_model = {}
        if self.model:
            prediction = self.classify_clip(
                clip,
                self.model,
                meta_data,
                reuse_frames=reuse_frames,
            )
            predictions_per_model[self.model.id] = prediction
        else:
            for model in self.config.classify.models:
                prediction = self.classify_clip(
                    clip,
                    model,
                    meta_data,
                    reuse_frames=reuse_frames,
                )
                predictions_per_model[model.id] = prediction
        destination_folder = os.path.dirname(filename)
        dirname = destination_folder

        if self.previewer:
            mpeg_filename = os.path.join(dirname, base_filename + "-classify.mp4")

            logging.info("Exporting preview to '{}'".format(mpeg_filename))

            self.previewer.export_clip_preview(
                mpeg_filename, clip, list(predictions_per_model.values())[0]
            )
        models = [self.model] if self.model else self.config.classify.models
        meta_data = self.save_metadata(
            meta_data,
            meta_file,
            clip,
            predictions_per_model,
            models,
            calculate_thumbnails=calculate_thumbnails,
        )
        if cache_to_disk:
            clip.frame_buffer.remove_cache()
        return meta_data

    def classify_clip(self, clip, model, meta_data, reuse_frames=None):
        start = time.time()
        classifier = self.get_classifier(model)
        predictions = Predictions(classifier.labels, model)
        predictions.model_load_time = time.time() - start

        for i, track in enumerate(clip.tracks):
            segment_frames = None
            if reuse_frames:
                tracks = meta_data.get("tracks")
                meta_track = next(
                    (x for x in tracks if x["id"] == track.get_id()), None
                )
                if meta_track is not None:
                    prediction_tag = next(
                        (
                            x
                            for x in meta_track["tags"]
                            if x.get("data", {}).get("name") == model.name
                        ),
                        None,
                    )
                    if prediction_tag is not None:
                        if "prediction_frames" in prediction_tag["data"]:
                            logging.info("Reusing previous prediction frames %s", model)
                            segment_frames = prediction_tag["data"]["prediction_frames"]
                            segment_frames = np.uint16(segment_frames)

            prediction = classifier.classify_track(
                clip, track, segment_frames=segment_frames, min_segments=1
            )
            if prediction is not None:
                predictions.prediction_per_track[track.get_id()] = prediction
                description = prediction.description()
                logging.info(
                    "{} - [{}/{}] prediction: {}".format(
                        track.get_id(), i + 1, len(clip.tracks), description
                    )
                )
        if self.config.verbose:
            ms_per_frame = (
                (time.time() - start) * 1000 / max(1, len(clip.frame_buffer.frames))
            )
            logging.info("Took {:.1f}ms per frame".format(ms_per_frame))
        if classifier.TYPE == "Keras":
            tools.clear_session()
        del classifier
        gc.collect()

        return predictions

    def save_metadata(
        self,
        meta_data,
        meta_filename,
        clip,
        predictions_per_model,
        models,
        calculate_thumbnails=False,
    ):
        tracks = meta_data.get("tracks")
        for track in clip.tracks:
            meta_track = next((x for x in tracks if x["id"] == track.get_id()), None)
            if meta_track is None:
                logging.error(
                    "Got prediction for track which doesn't exist in metadata"
                )
                continue
            prediction_info = []
            for model_id, predictions in predictions_per_model.items():
                prediction = predictions.prediction_for(track.get_id())
                if prediction is None:
                    continue
                prediction_meta = prediction.get_metadata()
                prediction_meta["model_id"] = model_id
                prediction_info.append(prediction_meta)
            meta_track["predictions"] = prediction_info

            if calculate_thumbnails:
                best_thumb, best_score = get_thumbnail_info(clip, track)
                if best_thumb is None:
                    meta_track["thumbnail"] = None
                else:
                    thumbnail_info = {
                        "region": best_thumb.region,
                        "contours": best_thumb.contours,
                        "median_diff": best_thumb.median_diff,
                        "score": round(best_score),
                    }
                    meta_track["thumbnail"] = thumbnail_info
        if calculate_thumbnails and len(clip.tracks) == 0:
            # if no tracks choose a clip thumb
            region = best_trackless_thumb(clip)
            meta_data["thumbnail_region"] = region

        model_dictionaries = []
        for model in models:
            model_dic = model.as_dict()
            model_predictions = predictions_per_model[model.id]
            model_dic["classify_time"] = round(
                model_predictions.classify_time + model_predictions.model_load_time, 1
            )
            model_dictionaries.append(model_dic)

        meta_data["models"] = model_dictionaries
        if self.config.classify.meta_to_stdout:
            logging.info("Printing json meta data")

            print(json.dumps(meta_data, cls=tools.CustomJSONEncoder))
        else:
            logging.info("saving meta data %s", meta_filename)
            with open(meta_filename, "w") as f:
                json.dump(meta_data, f, indent=4, cls=tools.CustomJSONEncoder)
        return meta_data



from datetime import datetime
import logging
import os
import yaml
from track.cliptrackextractor import ClipTrackExtractor
from cptv import CPTVWriter
from cptv import Frame
from datetime import timedelta
import time
from piclassifier.recorder import Recorder
from pathlib import Path
from ml_tools.logs import init_logging
import multiprocessing
from .eventreporter import log_event

CPTV_EXT = ".cptv"


class CPTVRecorder(Recorder):
    def __init__(
        self,
        thermal_config,
        headers,
        name="CPTVRecorder",
        **args,
    ):
        super().__init__(
            thermal_config,
            headers,
            name,
            CPTV_EXT,
            **args,
        )

    def new_recording(self, background_frame, preview_frames, temp_thresh, frame_time):
        can_rec = self.can_record(frame_time)
        if not can_rec:
            return False
        self.rec_p = multiprocessing.Process(
            target=record,
            args=(
                self.frame_q,
                self.filename,
                temp_thresh,
                self.preview_secs,
                self.motion,
                self.headers,
                self.location_config,
                self.device_config,
                background_frame,
                preview_frames,
            ),
        )
        self.rec_p.start()
        return True

    def final_name(self):
        return self.output_dir / self.filename.with_suffix(self.file_extention).name


def record(
    queue,
    filename,
    temp_thresh,
    preview_secs,
    motion_config,
    headers,
    location_config,
    device_config,
    background_frame=None,
    init_frames=None,
    name="CPTVRecorder",
):
    init_logging()
    frames = 0
    try:
        logging.info("%s Recorder %s started", name, filename.resolve())
        f = open(filename, "wb")
        writer = CPTVWriter(f)
        writer.timestamp = datetime.now()
        writer.latitude = location_config.latitude
        writer.longitude = location_config.longitude
        writer.preview_secs = preview_secs
        motion_config.temp_thresh = temp_thresh
        writer.motion_config = yaml.dump(motion_config.as_dict()).encode()[:255]
        if headers.model:
            writer.model = headers.model.encode()
        if headers.brand:
            writer.brand = headers.brand.encode()
        if headers.firmware:
            writer.firmware = headers.firmware.encode()
        writer.camera_serial = headers.serial
        if background_frame is not None:
            f = Frame(background_frame, timedelta(), timedelta(), 0, 0)
            f.background_frame = True
            writer.background_frame = f
        # add brand model fps etc to cptv when python-cptv supports
        if device_config.name:
            writer.device_name = device_config.name.encode()
        if device_config.device_id:
            writer.device_id = device_config.device_id

        writer.write_header()
        if init_frames is not None:
            for frame in init_frames:
                writer.write_frame(frame)
                frames += 1
        while True:
            frame = queue.get()
            if isinstance(frame, int) and frame == 0:
                writer.close()
                break
            writer.write_frame(frame)
            frames += 1

    except Exception as ex:
        logging.error("%s Error Recording %s", name, filename.resolve(), exc_info=True)
        log_event("error-recording", ex)
        try:
            writer.close()
        except:
            pass
    logging.info("%s Recorder %s Written %s", name, filename.resolve(), frames)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
r"""Convert raw COCO 2017 dataset to TFRecord.

Example usage:
    python create_coco_tf_record.py --logtostderr \
      --image_dir="${TRAIN_IMAGE_DIR}" \
      --image_info_file="${TRAIN_IMAGE_INFO_FILE}" \
      --object_annotations_file="${TRAIN_ANNOTATIONS_FILE}" \
      --caption_annotations_file="${CAPTION_ANNOTATIONS_FILE}" \
      --output_file_prefix="${OUTPUT_DIR/FILE_PREFIX}" \
      --num_shards=100
"""
import cv2
from PIL import Image
from pathlib import Path
import time
import collections
import hashlib
import io
import json
import os
from multiprocessing import Process, Queue

from absl import app
from absl import flags
from absl import logging
import numpy as np
from PIL import Image, ImageOps

import tensorflow as tf
from . import tfrecord_util
from ml_tools import tools
from ml_tools.imageprocessing import normalize
from ml_tools.forestmodel import forest_features
from ml_tools import imageprocessing
from ml_tools.frame import TrackChannels
from ml_tools.trackdatabase import TrackDatabase
from ml_tools.rawdb import RawDatabase
from ml_tools.rectangle import Rectangle

crop_rectangle = Rectangle(0, 0, 640, 480)
from functools import lru_cache


def create_tf_example(sample, data, features, labels, num_frames, country_code):
    """Converts image and annotations to a tf.Example proto.

    Args:
      image: dict with keys: [u'license', u'file_name', u'coco_url', u'height',
        u'width', u'date_captured', u'flickr_url', u'id']
      image_dir: directory containing the image files.
      bbox_annotations:
        list of dicts with keys: [u'segmentation', u'area', u'iscrowd',
          u'image_id', u'bbox', u'category_id', u'id'] Notice that bounding box
          coordinates in the official COCO dataset are given as [x, y, width,
          height] tuples using absolute coordinates where x, y represent the
          top-left (0-indexed) corner.  This function converts to the format
          expected by the Tensorflow Object Detection API (which is which is
          [ymin, xmin, ymax, xmax] with coordinates normalized relative to image
          size).
      category_index: a dict containing COCO category information keyed by the
        'id' field of each category.  See the label_map_util.create_category_index
        function.
      caption_annotations:
        list of dict with keys: [u'id', u'image_id', u'str'].
      include_masks: Whether to include instance segmentations masks
        (PNG encoded) in the result. default: False.

    Returns:
      example: The converted tf.Example
      num_annotations_skipped: Number of (invalid) annotations that were ignored.

    Raises:
      ValueError: if the image pointed to by data['filename'] is not a valid JPEG
    """
    average_dim = [r.area for r in sample.track_bounds]
    average_dim = int(round(np.mean(average_dim) ** 0.5))
    thermals = list(data[0])
    filtereds = list(data[1])
    image_id = sample.unique_id
    image_height, image_width = thermals[0].shape
    while len(thermals) < num_frames:
        # ensure 25 frames even if 0s
        thermals.append(np.zeros((thermals[0].shape)))
        filtereds.append(np.zeros((filtereds[0].shape)))
    thermals = np.array(thermals)
    filtereds = np.array(filtereds)
    thermal_key = hashlib.sha256(thermals).hexdigest()
    filtered_key = hashlib.sha256(filtereds).hexdigest()
    avg_mass = int(round(sample.mass / len(sample.frame_numbers)))
    feature_dict = {
        "image/filtered": tfrecord_util.int64_feature(1 if sample.filtered else 0),
        "image/avg_mass": tfrecord_util.int64_feature(avg_mass),
        "image/track_median_mass": tfrecord_util.int64_feature(
            int(sample.track_median_mass)
        ),
        "image/avg_dim": tfrecord_util.int64_feature(average_dim),
        "image/height": tfrecord_util.int64_feature(image_height),
        "image/width": tfrecord_util.int64_feature(image_width),
        "image/clip_id": tfrecord_util.int64_feature(sample.clip_id),
        "image/track_id": tfrecord_util.int64_feature(sample.track_id),
        "image/filename": tfrecord_util.bytes_feature(
            str(sample.source_file).encode("utf8")
        ),
        "image/source_id": tfrecord_util.bytes_feature(str(image_id).encode("utf8")),
        "image/thermalencoded": tfrecord_util.float_list_feature(thermals.ravel()),
        "image/filteredencoded": tfrecord_util.float_list_feature(filtereds.ravel()),
        "image/features": tfrecord_util.float_list_feature(features),
        "image/filteredkey/sha256": tfrecord_util.bytes_feature(
            filtered_key.encode("utf8")
        ),
        "image/thermalkey/sha256": tfrecord_util.bytes_feature(
            thermal_key.encode("utf8")
        ),
        "image/format": tfrecord_util.bytes_feature("jpeg".encode("utf8")),
        "image/class/text": tfrecord_util.bytes_feature(sample.label.encode("utf8")),
        "image/class/label": tfrecord_util.int64_feature(labels.index(sample.label)),
        "image/country_id": tfrecord_util.bytes_feature(
            str(country_code).encode("utf8")
        ),
    }

    example = tf.train.Example(features=tf.train.Features(feature=feature_dict))
    return example


# @lru_cache(maxsize=1000)
# only going to get called once now per track
def get_track_data(clip_id, track_id, db):
    background = db.get_clip_background(clip_id)

    track_frames = db.get_track(
        clip_id, track_id, channels=[TrackChannels.thermal], crop=True
    )
    clip_meta = db.get_clip_meta(clip_id)
    frame_temp_median = clip_meta["frame_temp_median"]
    regions = [f.region for f in track_frames]
    features = forest_features(
        track_frames, background, frame_temp_median, regions, normalize=True
    )
    return background, track_frames, features, frame_temp_median


def save_data(samples, writer, labels, extra_args):
    sample_data = get_data(samples, extra_args)
    if sample_data is None:
        return 0
    saved = 0
    try:
        country_code = sample_data[1]
        sample_data = sample_data[0]
        for sample, images, features in sample_data:
            tf_example = create_tf_example(
                sample, images, features, labels, extra_args["num_frames"], country_code
            )
            writer.write(tf_example.SerializeToString())
            saved += 1
    except:
        logging.error(
            "Could not save data for %s", samples[0].source_file, exc_info=True
        )
    return saved


def get_data(clip_samples, extra_args):
    # prepare the sample data for saving
    if len(clip_samples) == 0:
        return None
    data = []
    crop_rectangle = tools.Rectangle(2, 2, 160 - 2 * 2, 140 - 2 * 2)

    if clip_samples[0].source_file.suffix == ".hdf5":
        db = TrackDatabase(clip_samples[0].source_file)
    else:
        db = RawDatabase(clip_samples[0].source_file)
        db.load_frames()

    # going to redo segments to get rid of ffc segments
    clip_id = clip_samples[0].clip_id
    try:
        background = db.get_clip_background()
        if background is None:
            frame_data = db.get_frames()
            background = np.median(frame_data, axis=0)
            del frame_data
        clip_meta = db.get_clip_meta(extra_args.get("tag_precedence"))
        frame_temp_median = clip_meta.frame_temp_median

        # group samples by track_id
        samples_by_track = {}
        for s in clip_samples:
            samples_by_track.setdefault(s.track_id, []).append(s)

        for track_id in samples_by_track.keys():
            samples = samples_by_track[track_id]
            if clip_samples[0].source_file.suffix != ".hdf5":
                track = next(
                    (track for track in clip_meta.tracks if track.track_id == track_id),
                    None,
                )
                if extra_args.get("label_mapping") is not None:
                    track.remapped_label = extra_args["label_mapping"].get(
                        track.original_label, track.original_label
                    )
                if track is None:
                    logging.error(
                        "Cannot find track %s in clip %s", track_id, clip_meta.clip_id
                    )
                    continue
                # GP All assumes we dont have a track over multiple bins (Whcih we probably never want)
                if extra_args.get("use_segments", True):
                    track.get_segments(
                        segment_width=extra_args.get("segment_width", 25),
                        segment_frame_spacing=extra_args.get(
                            "segment_frame_spacing", 9
                        ),
                        segment_types=extra_args.get("segment_types"),
                        segment_min_mass=extra_args.get("segment_min_avg_mass"),
                        dont_filter=extra_args.get("dont_filter_segment", False),
                        skip_ffc=extra_args.get("skip_ffc", True),
                        ffc_frames=clip_meta.ffc_frames,
                        max_segments=len(samples),
                        frame_min_mass=extra_args.get("min_mass"),
                        filter_by_fp=extra_args.get("filter_by_fp"),
                    )
                else:
                    filter_by_lq = extra_args.get("filter_by_lq", False)
                    track.calculate_sample_frames(
                        min_mass=(
                            extra_args.get("min_mass")
                            if not filter_by_lq
                            else track.lower_mass
                        ),
                        max_mass=(
                            extra_args.get("max_mass")
                            if not filter_by_lq
                            else track.upper_mass
                        ),
                        ffc_frames=clip_meta.ffc_frames,
                        max_frames=extra_args.get("max_frames"),
                    )
                samples = track.samples
                frame_temp_median = {}
                track_frames = []
                for frame_i in range(
                    track.start_frame, track.start_frame + track.num_frames
                ):
                    f = db.frames[frame_i]
                    region = track.regions_by_frame[frame_i]
                    frame_temp_median[frame_i] = np.median(f.thermal)
                    cropped = f.crop_by_region(region)
                    cropped.float_arrays()
                    track_frames.append(cropped)

            else:
                track_frames = db.get_track(
                    clip_id, track_id, channels=[TrackChannels.thermal], crop=True
                )
            logging.debug("Saving %s samples %s", track_id, len(samples))
            used_frames = []

            features, _, _ = forest_features(
                track_frames,
                background,
                frame_temp_median,
                [f.region for f in track_frames],
                normalize=True,
                cropped=True,
            )

            by_frame_number = {}
            thermal_max_diff = None
            thermal_min_diff = None
            max_diff = None
            min_diff = None

            thermal_diff_norm = extra_args.get("thermal_diff_norm", False)

            for f in track_frames:
                if f.region.blank or f.region.width <= 0 or f.region.height <= 0:
                    continue

                by_frame_number[f.frame_number] = (f, frame_temp_median[f.frame_number])
                f.float_arrays()
                diff_frame = f.filtered
                new_max = np.amax(diff_frame)
                new_min = np.amin(diff_frame)
                if min_diff is None or new_min < min_diff:
                    min_diff = new_min
                    # min_diff = max(0, new_min)
                if max_diff is None or new_max > max_diff:
                    max_diff = new_max
                if thermal_diff_norm:
                    diff_frame = f.thermal - frame_temp_median[f.frame_number]
                    new_max = np.amax(diff_frame)
                    new_min = np.amin(diff_frame)
                    if thermal_min_diff is None or new_min < thermal_min_diff:
                        thermal_min_diff = new_min
                    if thermal_max_diff is None or new_max > thermal_max_diff:
                        thermal_max_diff = new_max

            # normalize by maximum difference between background and tracked region
            # probably only need to use difference on the frames used for this record
            # also min_diff maybe could just be set to 0 and clip values below 0,
            # these represent pixels whcih are cooler than the background
            for sample in samples:
                thermals = []  # np.empty(len(frames), dtype=object)
                filtered = []  # np.empty(len(frames), dtype=object)
                for frame_number in sample.frame_indices:
                    frame, temp_median = by_frame_number[frame_number]
                    # no need to do work twice
                    if frame_number not in used_frames:
                        used_frames.append(frame_number)
                        # frame.filtered = frame.thermal - frame.region.subimage(
                        #     background
                        # )
                        frame.resize_with_aspect(
                            (32, 32), crop_rectangle, keep_edge=True
                        )
                        if (
                            np.amax(frame.thermal) > 50000
                            or np.amin(frame.thermal) < 1000
                        ):
                            logging.error(
                                "Strange values for %s max %s min %s",
                                clip_id,
                                np.amax(frame.thermal),
                                np.amin(frame.thermal),
                            )
                            raise Exception(
                                f"Strange values for {clip_id} - {track_id} #{frame_number}"
                            )

                        frame.thermal -= temp_median
                        if not thermal_diff_norm:
                            np.clip(
                                frame.thermal, a_min=0, a_max=None, out=frame.thermal
                            )
                        frame.thermal, stats = imageprocessing.normalize(
                            frame.thermal,
                            min=thermal_min_diff,
                            max=thermal_max_diff,
                            new_max=255,
                        )
                        if not stats[0]:
                            frame.thermal = np.zeros((frame.thermal.shape))

                        frame.filtered, stats = imageprocessing.normalize(
                            frame.filtered, min=min_diff, max=max_diff, new_max=255
                        )
                        np.clip(frame.filtered, a_min=0, a_max=None, out=frame.filtered)

                        if not stats[0]:
                            frame.filtered = np.zeros((frame.filtered.shape))
                        f2 = np.uint8(frame.filtered)
                    filtered.append(frame.filtered)
                    thermals.append(frame.thermal)

                thermals = np.array(thermals)
                filtered = np.array(filtered)
                data.append((sample, (thermals, filtered), features))
    except:
        logging.error(
            "Cant get Samples for %s", clip_samples[0].source_file, exc_info=True
        )
        return None
    return (data, clip_meta.country_code)



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import attr

from .defaultconfig import DefaultConfig
from pathlib import Path


@attr.s
class TrainConfig(DefaultConfig):
    hyper_params = attr.ib()
    train_dir = attr.ib()
    epochs = attr.ib()
    resnet_params = attr.ib()
    use_gru = attr.ib()
    label_probabilities = attr.ib()
    type = attr.ib()
    LABEL_PROBABILITIES = {
        "bird": 20,
        "possum": 20,
        "rodent": 20,
        "hedgehog": 20,
        "cat": 5,
        "insect": 1,
        "leporidae": 5,
        "mustelid": 5,
        "false-positive": 1,
        "wallaby": 5,
        "vehicle": 1,
        "human": 1,
    }

    @classmethod
    def load(cls, raw, base_data_folder):
        resent_config = None
        if raw.get("resnet_params"):
            resent_config = ResnetConfig.load(raw.get("resnet_params"))
        return cls(
            type=raw["type"],
            resnet_params=resent_config,
            hyper_params=raw["hyper_params"],
            train_dir=Path(base_data_folder) / raw.get("train_dir", "train"),
            epochs=raw["epochs"],
            use_gru=raw["use_gru"],
            label_probabilities=raw["label_probabilities"],
        )

    @classmethod
    def get_defaults(cls):
        return cls(
            type="thermal",
            hyper_params={},
            resnet_params=None,
            train_dir=Path("train"),
            epochs=60,
            use_gru=True,
            label_probabilities=TrainConfig.LABEL_PROBABILITIES,
        )

    def validate(self):
        return True


@attr.s
class ResnetConfig:
    #  resnet
    num_filters = attr.ib()
    kernel_size = attr.ib()
    conv_stride = attr.ib()
    block_sizes = attr.ib()
    block_strides = attr.ib()
    bottleneck = attr.ib()
    resnet_size = attr.ib()
    first_pool_size = attr.ib()
    first_pool_stride = attr.ib()

    @classmethod
    def load(cls, raw):
        return cls(
            num_filters=raw.get("num_filters"),
            kernel_size=raw.get("kernel_size"),
            conv_stride=raw.get("conv_stride"),
            block_sizes=raw.get("block_sizes"),
            block_strides=raw.get("block_strides", [1, 2, 2, 2]),
            bottleneck=raw.get("bottleneck"),
            resnet_size=raw.get("resnet_size"),
            first_pool_size=raw.get("first_pool_size"),
            first_pool_stride=raw.get("first_pool_stride"),
        )



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from ml_tools.framecache import FrameCache
from ml_tools.frame import Frame
from ml_tools.tools import get_optical_flow_function
from threading import Lock


class FrameBuffer:
    """Stores entire clip in memory, required for some operations such as track exporting."""

    def __init__(
        self,
        cptv_name,
        high_quality_flow,
        cache_to_disk,
        calc_flow,
        keep_frames,
        max_frames=None,
    ):
        self.cache = FrameCache(cptv_name) if cache_to_disk else None
        self.opt_flow = None
        self.high_quality_flow = high_quality_flow
        self.frames = None
        self.prev_frame = None
        self.calc_flow = calc_flow
        self.max_frames = max_frames
        self.keep_frames = True if max_frames and max_frames > 0 else keep_frames
        self.current_frame_i = 0
        self.current_frame = None
        self.frame_lock = Lock()
        if calc_flow:
            self.set_optical_flow()
        self.reset()

    def set_optical_flow(self):
        if self.opt_flow is None:
            self.opt_flow = get_optical_flow_function(self.high_quality_flow)

    def add_frame(self, thermal, filtered, mask, frame_number, ffc_affected=False):
        self.prev_frame = self.current_frame
        frame = Frame(thermal, filtered, mask, frame_number, ffc_affected=ffc_affected)
        self.current_frame = frame

        if self.opt_flow:
            frame.generate_optical_flow(self.opt_flow, self.prev_frame)
        if self.keep_frames:
            if self.cache:
                self.cache.add_frame(frame)
            else:
                if self.max_frames and len(self.frames) == self.max_frames:
                    with self.frame_lock:
                        del self.frames[0]
                self.frames.append(frame)
        return frame

    @property
    def has_flow(self):
        return self.cache or self.opt_flow

    def get_frame(self, frame_number):
        if self.prev_frame and self.prev_frame.frame_number == frame_number:
            return self.prev_frame
        elif self.current_frame and self.current_frame.frame_number == frame_number:
            return self.current_frame
        elif self.cache:
            return self.cache.get_frame(frame_number)
        if self.current_frame is None:
            return None
        if frame_number > self.current_frame.frame_number:
            return None
        # this supports max frames etc
        frame_ago = self.current_frame.frame_number - frame_number
        frame = self.get_frame_ago(frame_ago)
        assert frame == None or frame.frame_number == frame_number
        return frame

    def close_cache(self):
        if self.cache:
            self.cache.close()

    def remove_cache(self):
        if self.cache:
            self.cache.delete()

    def get_frame_ago(self, n=1):
        if n == 0:
            return self.current_frame
        elif n == 1:
            return self.get_last_frame()
        if len(self.frames) > n:
            return self.frames[-(n + 1)]
        return None

    def get_last_frame(self):
        if self.prev_frame:
            return self.prev_frame
        elif len(self.frames) > 0:
            return self.frames[-1]
        return None

    def get_last_x(self, x=25):
        if self.cache and self.prev_frame:
            return self.prev_frame
        elif len(self.frames) > 0:
            return self.frames[-x:]
        return None

    def get_last_filtered(self, region=None):
        if self.cache:
            if not self.prev_frame:
                return None
            prev = self.prev_frame.filtered
        else:
            if len(self.frames) > 0:
                prev = self.frames[-1].filtered
            else:
                return None

        if region:
            return region.subimage(prev)
        else:
            return prev

    def reset(self):
        """
        Empties buffer
        """
        self.frames = []

    def __len__(self):
        return len(self.frames)

    def __iter__(self):
        if self.cache:
            self.cache.open(mode="r")
        return self

    def __next__(self):
        frame = self.get_frame(self.current_frame_i)
        if frame is None:
            raise StopIteration

        self.current_frame_i += 1
        return frame



import cv2
import numpy as np

from pathlib import Path
from PIL import Image
from scipy import ndimage
from PIL import Image
import logging


def resize_and_pad(
    frame,
    new_dim,
    region,
    crop_region,
    keep_edge=False,
    pad=None,
    interpolation=cv2.INTER_LINEAR,
    extra_h=0,
    extra_v=0,
):
    scale_percent = (new_dim[:2] / np.array(frame.shape[:2])).min()
    width = round(frame.shape[1] * scale_percent)
    height = round(frame.shape[0] * scale_percent)
    width = max(width, 1)
    height = max(height, 1)

    width = min(width, new_dim[0])
    height = min(height, new_dim[1])

    if len(frame.shape) == 3:
        resize_dim = (width, height, frame.shape[2])
    else:
        resize_dim = (width, height)
    if pad is None:
        pad = np.min(frame)

    resized = np.full(new_dim, pad, dtype=frame.dtype)
    offset_x = 0
    offset_y = 0
    frame_resized = resize_cv(frame, resize_dim, interpolation=interpolation)
    frame_height, frame_width = frame_resized.shape[:2]
    offset_x = (new_dim[1] - frame_width) // 2
    offset_y = (new_dim[0] - frame_height) // 2
    if keep_edge and crop_region is not None:
        if region.left <= crop_region.left:
            offset_x = 0
        elif region.right >= crop_region.right:
            offset_x = new_dim[1] - frame_width

        if region.top <= crop_region.top:
            offset_y = 0

        elif region.bottom >= crop_region.bottom:
            offset_y = new_dim[0] - frame_height
    if len(resized.shape) == 3:
        resized[
            offset_y : offset_y + frame_height, offset_x : offset_x + frame_width, :
        ] = frame_resized
    else:
        resized[
            offset_y : offset_y + frame_height,
            offset_x : offset_x + frame_width,
        ] = frame_resized

    return resized


def rotate(image, degrees, mode="nearest", order=1):
    return ndimage.rotate(image, degrees, reshape=False, mode=mode, order=order)


def resize_cv(image, dim, interpolation=cv2.INTER_LINEAR, extra_h=0, extra_v=0):
    return cv2.resize(
        np.float32(image),
        dsize=(dim[0] + extra_h, dim[1] + extra_v),
        interpolation=interpolation,
    )


def square_clip(data, frames_per_row, tile_dim, frame_samples, normalize=True):
    # lay each frame out side by side in rows
    new_frame = np.zeros((frames_per_row * tile_dim[0], frames_per_row * tile_dim[1]))
    i = 0
    success = False
    for x in range(frames_per_row):
        for y in range(frames_per_row):
            frame = data[frame_samples[i]]
            if normalize:
                frame, stats = normalize(frame, new_max=255)
                if not stats[0]:
                    continue
            success = True
            new_frame[
                x * tile_dim[0] : (x + 1) * tile_dim[0],
                y * tile_dim[1] : (y + 1) * tile_dim[1],
            ] = np.float32(frame)
            i += 1

    return new_frame, success


def square_clip_flow(data_flow, frames_per_row, tile_dim, use_rgb=False):
    if use_rgb:
        new_frame = np.zeros(
            (frames_per_row * tile_dim[0], frames_per_row * tile_dim[1], 3)
        )
    else:
        new_frame = np.zeros(
            (frames_per_row * tile_dim[0], frames_per_row * tile_dim[1])
        )

    i = 0
    success = False
    hsv = np.zeros((tile_dim[0], tile_dim[1], 3), dtype=np.float32)
    hsv[..., 1] = 255
    for x in range(frames_per_row):
        for y in range(frames_per_row):
            if i >= len(data_flow):
                flow = data_flow[-1]
            else:
                flow = data_flow[i]
            flow_h = flow[:, :, 0]
            flow_v = flow[:, :, 1]

            mag, ang = cv2.cartToPolar(flow_h, flow_v)
            hsv[..., 0] = ang * 180 / np.pi / 2
            hsv[..., 2] = cv2.normalize(mag, None, 0, 255, cv2.NORM_MINMAX)
            rgb = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)
            if use_rgb:
                flow_magnitude = rgb
            else:
                flow_magnitude = cv2.cvtColor(rgb, cv2.COLOR_BGR2GRAY)
            frame, norm_success = normalize(flow_magnitude)

            if not norm_success:
                continue
            success = True
            new_frame[
                x * tile_dim[0] : (x + 1) * tile_dim[0],
                y * tile_dim[1] : (y + 1) * tile_dim[1],
            ] = np.float32(frame)
            i += 1
    return new_frame, success


def normalize(data, min=None, max=None, new_max=1):
    """
    Normalize an array so that the values range from 0 -> new_max
    Returns normalized array, stats tuple (Success, min used, max used)
    """
    if data.size == 0:
        return np.zeros((data.shape)), (False, None, None)
    if max is None:
        max = np.amax(data)
    if min is None:
        min = np.amin(data)
    if max == min:
        if max == 0:
            return np.zeros((data.shape)), (False, max, min)
        data = data / max
        return data, (True, max, min)

    data = new_max * (np.float32(data) - min) / (max - min)
    return data, (True, max, min)


def save_image_channels(data, filename):
    Path(filename).parent.mkdir(parents=True, exist_ok=True)
    r = Image.fromarray(np.uint8(data[:, :, 0] * 255))
    g = Image.fromarray(np.uint8(data[:, :, 1] * 255))
    b = Image.fromarray(np.uint8(data[:, :, 2] * 255))
    concat = np.concatenate((r, g, b), axis=1)
    img = Image.fromarray(np.uint8(concat))
    img.save(filename + ".png")


index = 0


def detect_objects_ir(image, otsus=False, threshold=100, kernel=(15, 15)):
    image = np.uint8(image)
    # image = cv2.fastNlMeansDenoising(np.uint8(image), None)

    image = cv2.morphologyEx(image, cv2.MORPH_OPEN, kernel)

    # image = cv2.GaussianBlur(image, kernel, 0)
    flags = cv2.THRESH_BINARY
    if otsus:
        flags += cv2.THRESH_OTSU

    _, image = cv2.threshold(image, threshold, 255, flags)

    components, small_mask, stats, _ = cv2.connectedComponentsWithStats(image)
    return components, small_mask, stats


def detect_objects_both(
    salicencyMap, backsub, threshold=30, kernel=(15, 15), otsus=False
):
    if salicencyMap is not None:
        salicencyMap = np.uint8(salicencyMap)
        # image = cv2.fastNlMeansDenoising(np.uint8(image), None)

        salicencyMap = cv2.morphologyEx(salicencyMap, cv2.MORPH_OPEN, kernel)

        flags = cv2.THRESH_BINARY
        if otsus:
            flags += cv2.THRESH_OTSU

        _, salicencyMap = cv2.threshold(salicencyMap, threshold, 255, flags)

    backsub = np.uint8(backsub)
    backsub = cv2.GaussianBlur(backsub, kernel, 0)
    flags = cv2.THRESH_BINARY
    if otsus:
        flags += cv2.THRESH_OTSU
    _, backsub = cv2.threshold(backsub, threshold, 255, flags)
    # cv2.imshow("theshold", image)
    backsub = cv2.dilate(backsub, kernel, iterations=1)

    backsub = cv2.morphologyEx(backsub, cv2.MORPH_CLOSE, kernel)
    # cv2.imshow("backsub.png", np.uint8(backsub))
    both = backsub
    if salicencyMap is not None:
        # cv2.imshow("salicencyMap.png", np.uint8(salicencyMap))
        both = backsub | salicencyMap
        # cv2.imshow("both.png", np.uint8(both))

    # cv2.waitKey(10)

    components, small_mask, stats, _ = cv2.connectedComponentsWithStats(both)
    return components, small_mask, stats


def detect_objects(image, otsus=False, threshold=30, kernel=(15, 15)):
    image = np.uint8(image)
    image = cv2.GaussianBlur(image, kernel, 0)
    flags = cv2.THRESH_BINARY
    if otsus:
        flags += cv2.THRESH_OTSU
    _, image = cv2.threshold(image, threshold, 255, flags)
    image = cv2.morphologyEx(image, cv2.MORPH_CLOSE, kernel)
    return cv2.connectedComponentsWithStats(image)


def clear_frame(frame):
    filtered = frame.filtered
    thermal = frame.thermal
    if len(filtered) == 0 or len(thermal) == 0:
        return False
    thermal_deviation = np.amax(thermal) != np.amin(thermal)
    filtered_deviation = np.amax(filtered) != np.amin(filtered)
    if not thermal_deviation or not filtered_deviation:
        return False

    return True


def hist_diff(region, background, thermal, normalize_images=False):
    track_back = region.subimage(background).copy()
    track_thermal = region.subimage(thermal).copy()
    if normalize_images:
        track_back, _ = normalize(track_back, new_max=255)
        track_thermal, _ = normalize(track_thermal, new_max=255)
        track_back = np.float32(track_back)
        track_thermal = np.float32(track_thermal)
    h_bins = 60
    # s_bins = 60
    histSize = [h_bins]

    hist_base = cv2.calcHist(
        [track_back],
        None,
        None,
        histSize,
        [0, 255],
        accumulate=False,
    )
    cv2.normalize(hist_base, hist_base, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX)
    hist_track = cv2.calcHist(
        [track_thermal],
        None,
        None,
        histSize,
        [0, 255],
        accumulate=False,
    )

    cv2.normalize(
        hist_track,
        hist_track,
        alpha=0,
        beta=1,
        norm_type=cv2.NORM_MINMAX,
    )
    compared_v = cv2.compareHist(hist_track, hist_base, 0)
    return compared_v



from abc import ABC, abstractmethod
from datetime import datetime
import logging
import multiprocessing
import shutil
import time
from pathlib import Path

from ml_tools.logs import init_logging

TEMP_DIR = "temp"


class Recorder(ABC):
    def __init__(
        self,
        thermal_config,
        headers,
        name,
        file_extention,
        constant_recorder=False,
        on_recording_stopping=None,
        file_suffix=None,
    ):

        self.file_suffix = file_suffix
        self.file_extention = file_extention
        self.name = name
        self.constant_recorder = constant_recorder
        self.location_config = thermal_config.location
        self.device_config = thermal_config.device
        self.output_dir = Path(thermal_config.recorder.output_dir)
        if constant_recorder:
            self.output_dir = self.output_dir / "constant-recordings"
            self.output_dir.mkdir(parents=True, exist_ok=True)

        self.temp_dir = self.output_dir / TEMP_DIR
        self.temp_dir.mkdir(parents=True, exist_ok=True)
        self.motion = thermal_config.motion
        self.preview_secs = thermal_config.recorder.preview_secs
        self.filename = None
        self.recording = False
        self.frames = 0
        self.headers = headers
        self.min_disk_space_mb = thermal_config.recorder.min_disk_space_mb
        self.min_frames = thermal_config.recorder.min_secs * headers.fps
        self.max_frames = thermal_config.recorder.max_secs * headers.fps
        self.min_recording = self.preview_secs * headers.fps + self.min_frames
        self.write_until = 0
        self.rec_time = 0
        self.on_recording_stopping = on_recording_stopping
        self.frame_q = multiprocessing.Queue()
        self.rec_p = None

    def process_frame(self, movement_detected, cptv_frame, received_at):
        if self.recording:
            self.write_frame(cptv_frame)
            if movement_detected:
                self.write_until = self.frames + self.min_frames
            elif self.has_minimum():
                self.stop_recording(received_at)
                return

            if self.frames == self.max_frames:
                self.stop_recording(received_at)

    def has_minimum(self):
        return self.frames >= self.write_until

    def write_frame(self, frame):
        start = time.time()
        self.frame_q.put(frame)
        # self.writer.next_frame(frame)
        self.frames += 1
        self.rec_time += time.time() - start

    def can_record(self, frame_time):
        _, _, free = shutil.disk_usage(self.output_dir)
        free = free * 0.000001
        if free <= self.min_disk_space_mb:
            logging.warn(
                "%s cannot record as only have %s MB and need %s MB",
                self.name,
                free,
                self.min_disk_space_mb,
            )
        return free > self.min_disk_space_mb

    def force_stop(self):
        if not self.recording:
            return
        if self.frames > self.min_recording:
            self.stop_recording(time.time())
        else:
            logging.info(
                "%s Recording stopped early deleting short recording", self.name
            )
            self.delete_recording()

    def delete_recording(self):
        if self.recording:
            self.frame_q.put(0)
            self.rec_p.join()
            self.frame_q = multiprocessing.Queue()
            self.rec_p = None
            self.recording = False
        self.filename.unlink()

    def stop_recording(self, frame_time):
        start = time.time()
        self.rec_time += time.time() - start
        self.recording = False
        final_name = self.final_name()
        logging.info("%s Waiting for recorder to finish", self.name)
        self.frame_q.put(0)
        self.rec_p.join()
        self.frame_q = multiprocessing.Queue()
        self.rec_p = None
        logging.info(
            "%s recording %s ended %s frames %s time recording %s per frame ",
            self.name,
            final_name,
            self.frames,
            self.rec_time,
            self.rec_time / self.frames if self.frames > 0 else 0,
        )
        self.rec_time = 0
        self.write_until = 0
        # write metadata first
        logging.info("Stopped %s", final_name)
        if self.on_recording_stopping is not None:
            self.on_recording_stopping(final_name)

        self.filename.rename(final_name)

    def delete_excess(self):
        stat = shutil.disk_usage(self.output_dir)
        free_percent = stat[2] / stat[0]
        if free_percent >= 0.3:
            return
        recordings = list(self.output_dir.glob(f"*{self.file_extention}"))
        recordings.sort()
        if len(recordings) == 0:
            return
        while free_percent < 0.3 and len(recordings) > 0:
            logging.info("Deleting %s", recordings[0])
            recordings[0].unlink()
            meta = recordings[0].with_suffix(".txt")
            if meta.exists:
                logging.info("Deleting %s", meta)
                meta.unlink()
            recordings = recordings[1:]
            stat = shutil.disk_usage(self.output_dir)
            free_percent = stat[2] / stat[0]

    def start_recording(
        self, background_frame, preview_frames, temp_thresh, frame_time
    ):
        if self.constant_recorder:
            self.delete_excess()

        start = time.time()
        if self.recording:
            logging.warn("%s Already recording, stop recording first", self.name)
            return False

        self.frames = 0

        self.filename = self.new_temp_name(frame_time)
        started = self.new_recording(
            background_frame, preview_frames, temp_thresh, frame_time
        )
        if not started:
            return False
        self.rec_time = time.time() - start
        self.write_until = self.frames + self.min_frames
        self.recording = True

        logging.info(
            "%s recording %s started temp_thresh: %s",
            self.name,
            self.filename,
            temp_thresh,
        )

        return True

    def new_temp_name(self, frame_time):
        file_name = datetime.fromtimestamp(frame_time).strftime("%Y%m%d-%H%M%S.%f")
        if self.file_suffix is not None:
            file_name = f"{file_name}{self.file_suffix}"
        return self.temp_dir / f"{file_name}{self.file_extention}"

    @abstractmethod
    def new_recording(
        self, background_frame, preview_frames, temp_thresh, frame_time
    ): ...

    @abstractmethod
    def final_name(self): ...



"""
Helper functions for classification of the tracks extracted from CPTV videos
"""

import os.path
import numpy as np
import pickle
import json
import datetime
import glob
import cv2
import enum
import timezonefinder
from PIL import Image, ImageFont, ImageDraw
from pathlib import Path
from ml_tools.rectangle import Rectangle
from dateutil import parser
from enum import Enum

EPISON = 1e-5

LOCAL_RESOURCES = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources")
GLOBAL_RESOURCES = "/usr/lib/classifier-pipeline/resources"


class FrameTypes(enum.Enum):
    """Types of frames"""

    thermal_tiled = 0
    filtered_tiled = 1
    flow_tiled = 2
    overlay = 3
    flow_rgb = 4
    thermal = 5
    filtered = 6

    @staticmethod
    def is_valid(name):
        return name in FrameTypes.__members__.keys()


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, np.ndarray):
            return list(obj)
        elif isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, Rectangle):
            return obj.meta_dictionary()
        elif isinstance(obj, Path):
            return str(obj)
        elif isinstance(obj, Enum):
            return str(obj.name)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def load_colourmap(filename):
    with open(filename, "rb") as f:
        return pickle.load(f)


def convert_heat_to_img(frame, colormap=None, temp_min=None, temp_max=None):
    """
    Converts a frame in float32 format to a PIL image in in uint8 format.
    :param frame: the numpy frame contining heat values to convert
    :param colormap: an optional colormap to use, if none is provided then tracker.colormap is used.
    :return: a pillow Image containing a colorised heatmap
    """
    # normalise
    if colormap is None:
        colormap = _load_colourmap(None)
    if temp_min is None:
        temp_min = np.amin(frame)
    if temp_max is None:
        temp_max = np.amax(frame)
    frame = np.float32(frame)
    frame = (frame - temp_min) / (temp_max - temp_min)
    colorized = np.uint8(255.0 * colormap(frame))
    img = Image.fromarray(colorized[:, :, :3])  # ignore alpha
    return img


def load_clip_metadata(filename):
    """
    Loads a metadata file for a clip.
    :param filename: full path and filename to meta file
    :return: returns the stats file
    """
    with open(filename, "r") as t:
        # add in some metadata stats
        meta = json.load(t)
    if meta.get("recordingDateTime"):
        meta["recordingDateTime"] = parser.parse(meta["recordingDateTime"])
    if meta.get("tracks") is None and meta.get("Tracks"):
        meta["tracks"] = meta["Tracks"]
    return meta


def clear_session():
    import tensorflow as tf

    tf.keras.backend.clear_session()


def calculate_variance(filtered, prev_filtered):
    """Calculates variance of filtered frame with previous frame"""
    if filtered.size == 0:
        return 0
    if prev_filtered is None:
        return
    delta_frame = np.abs(filtered - prev_filtered)
    return np.var(delta_frame)


def get_optical_flow_function(high_quality=False):
    opt_flow = cv2.optflow.createOptFlow_DualTVL1()
    opt_flow.setUseInitialFlow(True)
    if not high_quality:
        # see https://stackoverflow.com/questions/19309567/speeding-up-optical-flow-createoptflow-dualtvl1
        opt_flow.setTau(1 / 4)
        opt_flow.setScalesNumber(3)
        opt_flow.setWarpingsNumber(3)
        opt_flow.setScaleStep(0.5)
    return opt_flow


def frame_to_jpg(
    frame, filename, colourmap_file=None, f_min=None, f_max=None, img_fmt="PNG"
):
    Path(filename).parent.mkdir(parents=True, exist_ok=True)

    colourmap = _load_colourmap(colourmap_file)
    if f_min is None:
        f_min = np.amin(frame)
    if f_max is None:
        f_max = np.amax(frame)
    img = convert_heat_to_img(frame, colourmap, f_min, f_max)
    img.save(filename, img_fmt)


def _load_colourmap(colourmap_path):
    if colourmap_path is None or not os.path.exists(colourmap_path):
        colourmap_path = resource_path("colourmap.dat")
    return load_colourmap(colourmap_path)


def resource_path(name):
    for base in [LOCAL_RESOURCES, GLOBAL_RESOURCES]:
        p = os.path.join(base, name)
        if os.path.exists(p):
            return p
    raise OSError("unable to locate {} resource".format(name))


def add_heat_number(img, frame, scale):
    draw = ImageDraw.Draw(img)
    font = ImageFont.truetype(resource_path("Ubuntu-R.ttf"), 8)
    for y, row in enumerate(frame):
        if y % 4 == 0:
            min_v = np.amin(row)
            min_i = np.where(row == min_v)[0][0]
            max_v = np.amax(row)
            max_i = np.where(row == max_v)[0][0]
            # print("min is", min_v, "max is", max_v)
            draw.text((min_i * scale, y * scale), str(int(min_v)), (0, 0, 0), font=font)
            draw.text((max_i * scale, y * scale), str(int(max_v)), (0, 0, 0), font=font)
            # print("drawing the max at row", y, max, max_i * scale, y * scale)


gzip_compression = {"compression": "gzip"}


def eucl_distance_sq(first, second):
    first_sq = first[0] - second[0]
    first_sq = first_sq * first_sq
    second_sq = first[1] - second[1]
    second_sq = second_sq * second_sq

    return first_sq + second_sq


def get_clipped_flow(flow):
    return np.clip(flow * 256, -16000, 16000)


def saveclassify_image(data, filename):
    # saves image channels side by side, expected data to be values in the range of 0->1
    Path(filename).parent.mkdir(parents=True, exist_ok=True)
    r = Image.fromarray(np.uint8(data[:, :, 0]))
    _, _, channels = data.shape

    if channels == 1:
        g = r
    else:
        g = Image.fromarray(np.uint8(data[:, :, 1]))

    if channels == 2:
        b = r
    else:
        b = Image.fromarray(np.uint8(data[:, :, 2]))
    concat = np.concatenate((r, g, b), axis=1)  # horizontally
    img = Image.fromarray(np.uint8(concat))
    img.save(filename + ".png")


def get_timezone_str(lat, lng):
    tf = timezonefinder.TimezoneFinder()
    timezone_str = tf.certain_timezone_at(lat=lat, lng=lng)

    if timezone_str is None:
        timezone_str = "Pacific/Auckland"
    return timezone_str


def saveclassify_rgb(data, filename):
    Path(filename).parent.mkdir(parents=True, exist_ok=True)
    r = Image.fromarray(np.uint8(data))
    r.save(filename + ".png")


def purge(dir, pattern):
    for f in glob.glob(os.path.join(dir, pattern)):
        os.remove(f)



from abc import ABC, abstractmethod
from struct import unpack_from
import numpy as np
from cptv import Frame


class RawFrame(ABC):
    def __init__(self, headers):
        self.pix = None
        self.telemetry = None
        self.res_x = headers.res_x
        self.res_y = headers.res_y
        self.img_dtype = np.dtype("uint{}".format(headers.pixel_bits))
        self.received_at = None

    def parse(self, data):
        telemetry = self.parse_telemetry(data[: self.get_telemetry_size()])

        thermal_frame = np.frombuffer(
            data, dtype=self.img_dtype, offset=self.get_telemetry_size()
        ).reshape(self.res_y, self.res_x)
        f = Frame(
            thermal_frame.byteswap(),
            telemetry.time_on,
            telemetry.last_ffc_time,
            telemetry.fpa_temp,
            telemetry.fpa_temp_last_ffc,
        )
        f.ffc_imminent = telemetry.ffc_imminent
        f.ffc_status = telemetry.ffc_status
        return f

    @abstractmethod
    def get_telemetry_size(self): ...

    @abstractmethod
    def parse_telemetry(self, raw_bytes): ...


def get_uint16(raw, offset):
    return unpack_from(">H", raw, offset)[0]


def get_uint32(raw, offset):
    return (
        raw[offset + 1]
        | (raw[offset] << 8)
        | (raw[offset + 3] << 16)
        | (raw[offset + 2] << 24)
    )


def get_uint64(raw, offset):
    return (
        raw[offset + 1]
        | (raw[offset] << 8)
        | (raw[offset + 3] << 16)
        | (raw[offset + 2] << 24)
        | (raw[offset + 5] << 32)
        | (raw[offset + 4] << 40)
        | (raw[offset + 7] << 48)
        | (raw[offset + 6] << 56)
    )



from struct import unpack_from
from datetime import timedelta
from .rawframe import RawFrame, get_uint32, get_uint16, get_uint64
from piclassifier.telemetry import Telemetry

import struct


class Lepton3(RawFrame):
    VOSPI_DATA_SIZE = 160
    TELEMETRY_PACKET_COUNT = 4

    def __init__(self, headers):
        super().__init__(headers)

    def get_telemetry_size(self):
        return Lepton3.VOSPI_DATA_SIZE * Lepton3.TELEMETRY_PACKET_COUNT

    def parse_telemetry(self, raw_bytes):
        offset = 0
        revision = get_uint16(raw_bytes, offset)
        time_counter = get_uint32(raw_bytes, 2)
        status_bits = get_uint32(raw_bytes, 6)
        ffc_status = (status_bits >> 4) & 0b11
        offset = 2 + 4 + 4 + 16
        software_revision = get_uint64(raw_bytes, offset)  # /26
        offset += 8 + 6
        frame_counter = get_uint32(raw_bytes, offset)
        offset += 4

        frame_mean, fpa_temp_counts, fpa_temp = unpack_from(
            ">HHH", raw_bytes, offset=offset
        )
        offset += 2 * (1 + 1 + 1 + 4)
        fpa_temp_last_ffc = get_uint16(raw_bytes, offset)

        offset += 2
        time_counter_last_ffc = get_uint32(raw_bytes, offset)
        ffc_imminent = raw_bytes[637] > 0

        # 60th byte
        t = Telemetry()
        t.telemetry_revision = revision
        t.time_on = timedelta(milliseconds=time_counter)
        t.status_bits = status_bits
        t.software_revision = software_revision
        t.frame_counter = frame_counter
        t.frame_mean = frame_mean
        t.fpa_temp_counts = fpa_temp_counts
        t.fpa_temp = (fpa_temp - 27315.0) / 100
        t.fpa_temp_last_ffc = (fpa_temp_last_ffc - 27315.0) / 100
        t.last_ffc_time = timedelta(milliseconds=time_counter_last_ffc)
        t.ffc_imminent = ffc_imminent
        t.ffc_status = ffc_status
        return t

    # not used for anything at the moment,  has not been tested
    def raw_telemetry(self):
        telemetry_data = b""
        telemetry_data += pack_uint16(self.telemetry.telemetry_revision)
        telemetry_data += pack_uint32(self.telemetry.timeon.milliseconds)
        status_bits = self.telemetry.status_bits << 4
        telemetry_data += pack_uint32(status_bits)

        telemetry_data += pack_uint64(self.telemetry.software_revision)

        telemetry_data += pack_uint32(self.telemetry.frame_counter)

        telemetry_data += pack_uint16(self.telemetry.frame_mean)
        telemetry_data += pack_uint16(self.telemetry.fpa_temp_counts)
        telemetry_data += pack_uint16(self.telemetry.fpa_temp * 100 + 27315)

        telemetry_data += pack_uint16(self.telemetry.fpa_temp_last_ffc * 100 + 27315)
        telemetry_data += pack_uint32(self.telemetry.last_ffc_time.milliseconds)
        remaining = self.get_telemetry_size() - len(telemetry_data)
        telemetry_data += bytearray(remaining)

        return telemetry_data

    def raw_data(self):
        telemetry_bytes = self.raw_telemetry
        frame_bytes = self.pix.tobytes()
        return telemetry_bytes + frame_bytes


def pack_uint16(data):
    return struct.pack(">H", data)


def pack_uint64(data):
    return struct.pack(">L", data)


def pack_uint32(data):
    return struct.pack(">L", data)



import attr
import logging
import numpy as np
import time
from attrs import define, field

# uniform prior stats start with uniform distribution.  This is the safest bet, but means that
# it takes a while to make predictions.  When off the first prediction is used instead causing
# faster, but potentially more unstable predictions.
UNIFORM_PRIOR = False


class Predictions:
    def __init__(self, labels, model):
        self.labels = labels
        self.prediction_per_track = {}
        self.model = model
        self.model_load_time = None

    def get_or_create_prediction(self, track, keep_all=True):
        prediction = self.prediction_per_track.setdefault(
            track.get_id(),
            TrackPrediction(
                track.get_id(),
                self.labels,
                keep_all=keep_all,
                start_frame=track.start_frame,
            ),
        )
        return prediction

    def clear_predictions(self):
        self.prediction_per_track = {}

    def prediction_for(self, track_id):
        return self.prediction_per_track.get(track_id)

    def guesses_for(self, track_id):
        prediction = self.prediction_per_track.get(track_id)
        if prediction:
            return prediction.guesses()
        return []

    def print_predictions(self, track_id):
        self.prediction_for(track_id).print_prediction()

    def prediction_description(self, track_id):
        return self.prediction_per_track.get(track_id).description()

    @property
    def classify_time(self):
        classify_time = [
            prediction.classify_time
            for prediction in self.prediction_per_track.values()
            if prediction.classify_time is not None
        ]
        return np.sum(classify_time)


@define
class Prediction:
    prediction = field()
    smoothed_prediction = field()
    frames = field()
    predicted_at_frame = field()
    mass = field()
    predicted_time = field(init=False)

    def __attrs_post_init__(self):
        self.predicted_time = time.time()

    def get_metadata(self):
        meta = attr.asdict(self)
        meta["smoothed_prediction"] = np.uint32(np.round(self.smoothed_prediction))
        meta["prediction"] = np.uint8(np.round(100 * self.prediction))
        return meta

    def clarity(self):
        best = np.argsort(self.prediction)
        return self.prediction[best[-1]] - self.prediction[best[-2]]

    def __str__(self):
        return f"{self.frames} conf: {np.round(100*self.prediction)}"


class TrackPrediction:
    """
    Class to hold the information about the predicted class of a track.
    Predictions are recorded for every frame, and methods provided for extracting the final predicted class of the
    track.
    """

    def __init__(self, track_id, labels, keep_all=True, start_frame=None):
        try:
            fp_index = labels.index("false-positive")
        except ValueError:
            fp_index = None

        self.track_id = track_id
        self.predictions = []
        self.fp_index = fp_index
        self.class_best_score = np.zeros((len(labels)))
        self.start_frame = start_frame

        self.last_frame_classified = None
        self.num_frames_classified = 0
        self.keep_all = keep_all
        self.labels = labels
        self.classify_time = None
        self.tracking = False
        self.masses = []
        self.normalized = False

    def cap_confidences(self, max_confidence):
        max_score = np.sum(self.class_best_score)
        if max_score > max_confidence:
            scale = max_confidence / max_score
            self.class_best_score *= scale

    def classified_clip(
        self,
        predictions,
        smoothed_predictions,
        prediction_frames,
        masses,
        top_score=None,
    ):
        self.num_frames_classified = len(predictions)
        for prediction, smoothed_prediction, frames, mass in zip(
            predictions, smoothed_predictions, prediction_frames, masses
        ):
            prediction = Prediction(
                prediction,
                smoothed_prediction,
                frames,
                np.amax(frames),
                mass,
            )
            self.predictions.append(prediction)

        if self.num_frames_classified > 0:
            self.class_best_score = np.sum(smoothed_predictions, axis=0)
            # normalize so it sums to 1
            if top_score is None:
                self.class_best_score = self.class_best_score / np.sum(
                    self.class_best_score
                )
            else:
                # possibility it doesn't sum to 1 i.e multi label model
                self.class_best_score /= top_score

    def normalized_best_score(self):
        return self.class_best_score[self.best_label_index] / np.sum(
            self.class_best_score
        )

    def get_normalized_score(self):
        score = None
        if self.class_best_score is not None:
            score = self.class_best_score / np.sum(self.class_best_score)
        return score

    def normalize_score(self):
        # normalize so it sums to 1
        if self.class_best_score is not None:
            self.class_best_score = self.class_best_score / np.sum(
                self.class_best_score
            )
            self.normalized = True

    def classified_frames(self, frame_numbers, predictions, mass):
        self.num_frames_classified += len(frame_numbers)
        self.last_frame_classified = np.max(frame_numbers)
        smoothed_prediction = predictions**2 * mass
        prediction = Prediction(
            predictions,
            smoothed_prediction,
            frame_numbers,
            self.last_frame_classified,
            mass,
        )
        if self.keep_all:
            self.predictions.append(prediction)
        else:
            self.predictions = [prediction]

        if self.normalized:
            logging.warning("Already normalized and still adding predicitions")
        if self.class_best_score is None:
            self.class_best_score = smoothed_prediction.copy()
        else:
            self.class_best_score += smoothed_prediction

    def classified_frame(self, frame_number, predictions, mass):
        self.last_frame_classified = frame_number
        self.num_frames_classified += 1
        self.masses.append(mass)
        smoothed_prediction = predictions**2 * mass

        prediction = Prediction(
            predictions,
            smoothed_prediction,
            frame_number,
            self.last_frame_classified,
            mass,
        )
        if self.keep_all:
            self.predictions.append(prediction)
        else:
            self.predictions = [prediction]

        if self.normalized:
            logging.warning("Already normalized and still adding predicitions")
        if self.class_best_score is None:
            self.class_best_score = smoothed_prediction
        else:
            self.class_best_score += smoothed_prediction

    def get_priority(self, frame_number):
        if self.tracking:
            return 100
        if self.last_frame_classified:
            skipepd_frames = frame_number - self.last_frame_classified
        else:
            skipepd_frames = frame_number - self.start_frame

        priority = skipepd_frames / 9
        if self.num_frames_classified == 0:
            priority += 2
        if self.fp_index and self.best_label_index == self.fp_index:
            # dont bother with fps unless nothing else to do
            priority -= 100
        logging.debug(
            "priority {} for track# {} num_frames {} last classified {} skipped {}".format(
                priority,
                self.track_id,
                self.num_frames_classified,
                self.last_frame_classified,
                skipepd_frames,
            )
        )
        return priority

    def get_prediction(self):
        return self.description()

    def get_classified_footer(self, frame_number=None):
        if len(self.predictions) == 0 or not self.keep_all:
            return "no classification"
        score = self.score_at_time(frame_number) * 10
        label = self.labels[self.label_at_time(frame_number)]

        score_2 = self.score_at_time(frame_number, n=2) * 10
        label_2 = self.labels[self.label_at_time(frame_number, n=2)]
        footer = "({:.1f} {}) second guess ({:.1f} {})".format(
            score, label, score_2, label_2
        )
        return footer

    def get_result(self):
        if len(self.predictions) > 0:
            return TrackResult(
                self.labels[self.best_label_index],
                self.max_score,
            )
        else:
            return None

    def description(self):
        """
        Returns a summary description of this prediction
        :param classes: Name of class for each label.
        :return:
        """
        score = self.max_score
        if score is None:
            return None
        if score > 0.5:
            first_guess = "{} {:.1f} (clarity {:.1f})".format(
                self.labels[self.best_label_index], score * 10, self.clarity * 10
            )
        else:
            first_guess = "[nothing] {} {:.1f} (clarity {:.1f})".format(
                self.labels[self.best_label_index], score * 10, self.clarity * 10
            )

        second_score = self.score(2)

        if second_score > 0.5:
            second_guess = "[second guess - {} {:.1f}]".format(
                self.labels[self.label_index(2)], second_score * 10
            )
        else:
            second_guess = ""

        return (first_guess + " " + second_guess).strip()

    @property
    def num_frames(self):
        return self.num_frames_classified

    def predicted_tag(self):
        index = self.best_label_index
        if index is None:
            return None
        return self.labels[index]

    def class_confidences(self):
        confidences = {}
        if self.class_best_score is None:
            return confidences
        for i, value in enumerate(self.class_best_score):
            confidences[self.labels[i]] = round(float(value), 3)
        return confidences

    @property
    def best_label_index(self):
        if self.class_best_score is None:
            return None
        return np.argmax(self.class_best_score)

    @property
    def max_score(self):
        if self.class_best_score is None:
            return None

        return float(np.amax(self.class_best_score))

    def clarity_at(self, frame):
        return self.predictions[frame].clarity

    @property
    def clarity(self):
        """The distance between our highest scoring class and second highest scoring class."""

        if self.class_best_score is None or len(self.class_best_score) < 2:
            return None
        return self.max_score - self.score(2)

    def label_index(self, n=None):
        """index of label of nth best guess."""

        if n is None:
            return self.best_label_index
        if self.class_best_score is None:
            return None
        return int(np.argsort(self.class_best_score)[-n])

    def score(self, n=None):
        """class prediction of nth best guess."""
        if n is None:
            return self.max_score
        if self.class_best_score is None:
            return None
        return float(sorted(self.class_best_score)[-n])

    def label_at_time(self, frame_number, n=1):
        """class label of nth best guess at a point in time."""
        if n is None:
            return None
        frames_per_prediction = len(self.smoothed_predictions) / self.num_frames
        prediction_index = int(frame_number / frames_per_prediction) + 1
        average = np.mean(self.smoothed_predictions[:prediction_index], axis=0)
        return int(np.argsort(average)[-n])

    def label_at_time(self, frame_number, n=1):
        """class prediction of nth best at a point in time."""
        if n is None:
            return None
        predictions = []
        class_best_score = None
        for pred in self.predictions:
            a_min = np.amin(pred.frames)
            a_max = np.amax(pred.frames)

            if a_min <= frame_number:
                predictions.append(pred.smoothed_prediction)

        class_best_score = np.sum(predictions, axis=0)
        if len(predictions) == 0:
            return 0
        class_best_score = class_best_score / np.sum(class_best_score)
        return int(np.argsort(class_best_score)[-n])

    def score_at_time(self, frame_number, n=1):
        """class prediction of nth best at a point in time."""
        if n is None:
            return None
        predictions = []
        class_best_score = None
        for pred in self.predictions:

            a_min = np.amin(pred.frames)
            a_max = np.amax(pred.frames)

            if a_min <= frame_number:
                predictions.append(pred.smoothed_prediction)
        class_best_score = np.sum(predictions, axis=0)
        if len(predictions) == 0:
            return 0

        class_best_score = class_best_score / np.sum(class_best_score)
        return float(sorted(class_best_score)[-n])

    def print_prediction(self):
        logging.info(
            "Track {} prediction {}".format(self.track_id, self.get_classified_footer())
        )

    def guesses(self):
        guesses = [
            "{} ({:.1f})".format(self.labels[self.label_index(i)], self.score(i) * 10)
            for i in range(1, min(len(self.labels), 4))
            if self.score(i) and self.score(i) > 0.5
        ]
        return guesses

    def get_metadata(self):
        prediction_meta = {}
        if self.classify_time is not None:
            prediction_meta["classify_time"] = round(self.classify_time, 1)

        prediction_meta["label"] = self.predicted_tag()
        # GP makes api pick up the label this will change when logic is moved to API
        prediction_meta["confident_tag"] = self.predicted_tag()

        prediction_meta["confidence"] = (
            round(self.max_score, 2) if self.max_score else 0
        )
        prediction_meta["clarity"] = round(self.clarity, 3) if self.clarity else 0
        prediction_meta["all_class_confidences"] = {}
        preds = []
        for p in self.predictions:
            preds.append(p.get_metadata())
        prediction_meta["predictions"] = preds
        if self.class_best_score is not None:
            for i, value in enumerate(self.class_best_score):
                label = self.labels[i]
                prediction_meta["all_class_confidences"][label] = round(value, 3)
        return prediction_meta


@attr.s(slots=True)
class TrackResult:
    what = attr.ib()
    confidence = attr.ib()



import dbus
import logging
import numpy as np

NAME = "org.cacophony.beacon"
PATH = "/org/cacophony/beacon"

# becaons have been hard coded for old labels, so remapping new label order to send as was previously

label_remap = [0, 11, 1, 2, 3, 4, 12, 5, 6, 13, 7, 8, 9, 10]

#
# def label_mapping():
#     remap = np.arange(14)
#     new_index = 0
#     tack_on = ["bird/kiwi", "insect", "penguin"]
#     tack_on = 11
#     for i, l in enumerate(labels):
#         if l in tack_on:
#             remap[i] = tack_on
#             tack_on += 1
#         else:
#             remap[i] = new_index
#             new_index += 1


#
# def init_logging(timestamps=False):
#     """Set up logging for use by various classifier pipeline scripts.
#
#     Logs will go to stderr.
#     """
#
#     fmt = "%(process)d %(thread)s:%(levelname)7s %(message)s"
#     if timestamps:
#         fmt = "%(asctime)s " + fmt
#     logging.basicConfig(
#         stream=sys.stderr, level=logging.DEBUG, format=fmt, datefmt="%Y-%m-%d %H:%M:%S"
#     )
#
# #
# def main():
#
#     recording()
#     recording()


def recording():
    bus = dbus.SystemBus()
    try:
        beacon_dbus = bus.get_object(NAME, PATH)
        beacon_dbus.Recording()
    except:
        logging.error("Dbus beacon error ", exc_info=True)


def classification(predictions):
    byte_predictions = {}

    for track_prediction in predictions:
        logging.debug(
            "%s prediction %s",
            track_prediction.track_id,
            track_prediction.predicted_tag(),
        )
        if track_prediction.class_best_score is None:
            return
        class_best_score = (
            100
            * track_prediction.class_best_score
            / np.sum(track_prediction.class_best_score)
        )
        for cur_i, confidence in enumerate(class_best_score):
            i = label_remap[cur_i]
            if dbus.Byte(i) in byte_predictions:
                byte_predictions[dbus.Byte(i)] = max(
                    dbus.Byte(int(confidence)), byte_predictions[dbus.Byte(i)]
                )
            else:
                byte_predictions[dbus.Byte(i)] = dbus.Byte(int(confidence))

    logging.debug("Sending classifcations %s ", byte_predictions)
    bus = dbus.SystemBus()
    try:
        beacon_dbus = bus.get_object(NAME, PATH)
        beacon_dbus.Classification(byte_predictions)
    except:
        logging.error("Dbus beacon error ", exc_info=True)


#
# if __name__ == "__main__":
#     main()
